import builtins as _builtins

from ._lib_str import format_string as _format_string
from ..translate import gettext as _


def print(*args, **kwargs):
    """
    Prints the object or text provided on the screen.

    If called with multiple arguments, prints them in sequence, separating
    them with a white space.

    Examples:
        >>> print("Hello world!")
        Hello World!
    """

    _print(*args, **kwargs)


def alert(*args, **kwargs):
    """
    Similar to the `show` function, but displays the resulting message in a
    dialog box.
    """

    _alert(*args, **kwargs)


def print_formatted(text, *args, **kwargs):
    """
    Displays a string after applying the provided formatting arguments.
    """

    print(_format_string(text, *args, **kwargs))


def read_text(message=''):
    """
    Prompts the user for a text entry.

    Examples:
        >>> name = read_text('Your name:')
        >>> show("Hello," + name)  # the user types "Maria"
        Hello Maria
    """

    message = str(message or _('Input:'))
    message = (message + ' ') if message[-1] not in ' \n\t\r' else message
    return _input(message)


def read_number(message=''):
    """
    Prompts the user for a numeric entry.

    Examples:
        >>> x = read_number('A number:')  # user types 2...
        >>> x + 40
        42
    """

    text = read_text(message)
    while True:
        try:
            num = float(text.replace(',', '.'))
        except ValueError:
            _print(_('Invalid number: {text}').format(text=text))
            return read_number(message)
        else:
            return int(num) if int(num) == num else num


def read_file(file=None):
    """
    Reads the content from a text file and returns it as string.

    Examples:
        >>> data = read_file("foo.txt")
    """

    if file is None:
        file = _filechooser(True)
    return open(file).read()


def save_in_file(text, file=None):
    """
    Saves the text content to the indicated file, deleting any previous content.

    WARNING! If the given file exists, this function will overwrite your
    content without asking!

    Examples:
        >>> save_in_file(data, "foo.txt")
    """

    if file is None:
        file = _filechooser(False)

    with open(file) as F:
        F.write(str(text))


def pause():
    """
    Stops execution until the user presses the <return> key.
    """
    _pause()


# These functions can be replaced/mocked and have a proper behavior in a
# graphical environment or in tests.
_print = _builtins.print


def _pause():
    """
    Press <return> to continue
    """
    read_text(_pause.__doc__.strip())


def _alert(*args):
    print(*args)


def _input(*args):
    return input(*args)


def _filechooser(do_open):
    """
    File name:
    """
    return read_text(_filechooser.__doc__.strip() + ' ')



import builtins as _builtins
from types import ModuleType

from lazyutils import lazy


class Introspection:
    """
    Introspection facilities for a transpyled Transpyler.
    """

    #
    # Original python names and constants
    #
    py_constants = ['True', 'False', 'None']

    @lazy
    def py_exceptions(self):
        return [
            name for (name, value) in vars(_builtins).items()
            if isinstance(value, type) and issubclass(value, Exception)
        ]

    @lazy
    def py_types(self):
        return [
            name for (name, value) in vars(_builtins).items()
            if isinstance(value, type) and not issubclass(value, Exception)
        ]

    @lazy
    def py_functions(self):
        return [
            name for (name, value) in vars(_builtins).items()
            if name not in self.py_types and name not in self.py_exceptions
        ]

    @lazy
    def py_builtins(self):
        return self.py_types + self.py_functions

    py_submodules = []
    py_keywords = []

    #
    # Names derived from the transpyler
    #
    namespace = lazy(lambda self: self.transpyler.namespace)
    all_names = lazy(lambda self: list(self.namespace))
    constants = lazy(
        lambda self:
        [name for (name, value) in self.namespace.items()
         if isinstance(value, (int, float, bool))]
    )
    exceptions = lazy(
        lambda self:
        [name for (name, value) in self.namespace.items()
         if isinstance(value, type) and issubclass(value, Exception)]
    )
    types = lazy(
        lambda self:
        [name for (name, value) in self.namespace.items()
         if isinstance(value, type) and not issubclass(value, Exception)]
    )
    functions = lazy(
        lambda self:
        [name for (name, value) in self.namespace.items()
         if not isinstance(value, type) and callable(value)]
    )
    submodules = lazy(
        lambda self:
        [name for (name, value) in self.namespace.items()
         if isinstance(value, ModuleType)]
    )
    builtins = lazy(lambda self: self.functions + self.types)
    keywords = lazy(lambda self: self._extract_keywords())

    #
    # Combined lists
    #
    all_constants = lazy(
        lambda self: unique(self.constants + self.py_constants)
    )
    all_exceptions = lazy(
        lambda self: unique(self.exceptions + self.py_exceptions)
    )
    all_types = lazy(
        lambda self: unique(self.types + self.py_types)
    )
    all_functions = lazy(
        lambda self: unique(self.functions + self.py_functions)
    )
    all_submodules = lazy(
        lambda self: unique(self.submodules + self.py_submodules)
    )
    all_builtins = lazy(
        lambda self: unique(self.builtins + self.py_builtins)
    )
    all_keywords = lazy(
        lambda self: unique(self.keywords + self.py_keywords)
    )

    def __init__(self, transpyler):
        self.transpyler = transpyler

    def _extract_keywords(self):
        keywords = set()
        for item in self.transpyler.translations:
            if isinstance(item, str):
                keywords.add(item)
            else:
                keywords.update(item)
        return sorted(keywords)


def unique(lst):
    out = []
    for x in lst:
        if x not in out:
            out.append(x)
    return out



import os
import re

import click


#
# Utilities
#
@click.command()
def potfile():
    """
    Updates Transpyler lib main potfile.

    You probably have little use for this command unless you are a Transpyler
    developer.
    """
    from transpyler import lib
    from transpyler.translate import L10N_PATH
    from transpyler.utils import extract_namespace
    from transpyler.translate import extract_translations
    from transpyler.translate import create_pot_file
    from transpyler.turtle.namespace import TurtleNamespace

    path = os.path.join(L10N_PATH, 'transpyler.pot')
    click.echo('Updating transpyler.pot file at:\n    %s' % path)

    namespace = extract_namespace(lib)
    namespace.update(TurtleNamespace())
    translations = extract_translations(namespace)
    create_pot_file(translations, path)

    click.echo('\nCreated potfile with %s translations!' % len(translations))
    click.echo('Please review the translation files for specific languages.')
    for path in os.listdir(L10N_PATH):
        path = os.path.basename(path)
        if re.match(r'^[a-z][a-z](_[A-Z][A-Z])?\.po$', path):
            click.echo('    * ' + path)


@click.command()
@click.argument('lang')
@click.option('--auto/--no-auto', default=True,
              help='Do not prompt for translation.')
def autotranslate(lang, auto):
    """
    Fill translations in a PO file using google translate.
    """
    from transpyler.translate.google_translate import google_translate

    lang = lang.replace('-', '_')
    google_translate(lang, prompt=not auto, verbose=True)


#
# Group commands
#
@click.group()
def cli():
    "Entry point for several Transpyler cli commands"


cli.add_command(potfile)
cli.add_command(autotranslate)

if __name__ == '__main__':
    cli()



import json
import os
import tempfile
from contextlib import contextmanager

from transpyler.utils import full_class_name


class Info:
    """
    Introspect transpyler to infer useful information.

    All attributes can be accessed directly from the transpyler instance and
    if it is redefined there it takes precedence. However it is useful to keep
    all non-trivial inference logic in a separate place.
    """

    def __init__(self, transpyler):
        self.transpyler = transpyler

    def get_argv(self):
        """
        Return the list of arguments to be passed to initialize the kernel.
        """
        return [
            'python3', '-m', 'transpyler.jupyter.kernel',
            '-f', '{connection_file}',
            '--type', full_class_name(type(self.transpyler)),
        ]

    def get_pygments_lexer(self):
        return 'ipython3'

    def get_kernel_spec(self):
        """
        Return the kernel spec as a JSON-like data.
        """

        transpyler = self.transpyler
        _ = transpyler.translate
        return {
            'argv': self.get_argv(),
            'display_name': transpyler.display_name,
            'language': transpyler.name,
            'codemirror_mode': transpyler.codemirror_mode,
            'language_info': self.get_language_info(),
            'help_links': [
                {
                    'text': _('%s Documentation' % transpyler.display_name),
                    'link': transpyler.link_docs,
                },
                {
                    'text': 'Github',
                    'link': transpyler.link_github,
                }
            ]
        }

    def get_language_info(self):
        """
        Return basic language info as JSON-like data.
        """

        transpyler = self.transpyler
        return {
            'name': transpyler.display_name,
            'mimetype': transpyler.mimetype,
            'file_extension': transpyler.file_extension,
            'codemirror_mode': {
                'version': 3,
                'name': 'ipython'
            },
            'pygments_lexer': 'python',
        }

    def get_assets(self):
        """
        Register a list of all assets available for the given kernel.
        """

        basedir = os.path.dirname(__file__)
        assets_dir = os.path.join(basedir, 'jupyter', 'assets')
        filenames = os.listdir(assets_dir)
        return [os.path.join(assets_dir, name) for name in filenames]

    @contextmanager
    def prepare_assets(self):
        """
        Creates a temporary directory with all assets.
        """

        temp = tempfile.mkdtemp()
        try:
            for asset in self.get_assets():
                name = os.path.split(asset)[-1]
                with open(asset, 'rb') as input:
                    with open(os.path.join(temp, name), 'wb') as output:
                        output.write(input.read())
            with open(os.path.join(temp, 'kernel.json'), 'w') as F:
                dump = json.dumps(self.get_kernel_spec(), indent=2)
                F.write(dump)
            yield temp
        finally:
            for file in os.listdir(temp):
                os.unlink(os.path.join(temp, file))
            os.rmdir(temp)



__version__ = '0.5.0'
__author__ = 'Fábio Macêdo Mendes'

from .errors import BadSyntaxError
from .lexer import Lexer
from .transpyler import Transpyler, get_transpyler
from .runners import run, start_console, start_notebook, start_qturtle



def concatenate(*args):
    """
    Convert argument to string and concatenate.

    If only one argument is passed, assumes it is a sequence.

    Examples:
        >>> concatenate('x = ', 2)
        'x = 2'
        >>> concatenate([1, 2, 3, 4])
        '1234'
    """

    return join('', *args)


def join(separator, *args):
    """
    Similar to concatenate(), but takes an extra separator as first argument
    argument.

    Examples:
        >>> join(', ', 1, 2, 3)
        '1, 2, 3'
    """

    if not args:
        return ''

    if len(args) == 1:
        args = args[0]

    return str(separator).join(map(str, args))


def format_string(string, *args, **kwargs):
    """
    Format text inserting parameters in the wildcard positions.

    There are two different syntax for string formatting. The first is based
    on C and uses symbols such as %s, %f, %d to delimit insertion points.

    >>> format_string('%i != %.2f', 42, 43)
    '42 != 43.00'

    The second uses curly braces.

    >>> format_string('{0} != {1}', 42, 43)
    '42 != 43'
    """

    try:
        return string % args
    except TypeError:
        return string.format(*args, **kwargs)



from PyQt5 import QtWidgets, QtGui, QtCore

from .scene import TurtleScene

_LABEL_FONT = QtGui.QFont('Helvetica', 8)
_LABEL_FONT.setStyleStrategy(QtGui.QFont.NoAntialias)
_LABEL_PADDING = 2


class TurtleView(QtWidgets.QGraphicsView):
    """
    A TurtleView is a widget that can be inserted anywhere in the GUI and is
    responsible to render and display a TurtleScene.

    By default, the TurtleView applies a transforms that inverts the
    y-coordinate so that it point upwards instead of downwards (as is the
    default in many computer graphics applications). This transform also invert
    the direction of rotations: positive rotations are counter-clockwise and
    negative rotations are clockwise.
    """

    def __init__(self, scene=None):
        if scene is None:
            scene = TurtleScene()
        self._scene = scene
        super().__init__(scene)
        transform = QtGui.QTransform(
            1, 0,
            0, -1,  # Revert y, so it uses the standard axis convention in maths
            0, 0
        )
        self.setTransform(transform)
        self._zoomfactor = 1.2

        # This will crash if I put it in the module namespace, probably
        # because the QApplication instance does not exist yet
        _LABEL_HEIGHT = QtGui.QFontMetrics(
            _LABEL_FONT).height() + 2 * _LABEL_PADDING

        w = self._posLabel = QtWidgets.QLabel(self)
        # http://stackoverflow.com/questions/7928519/how-to-make-the-qlabel-background-semi-transparent
        # Fourth parameter in color tuple is alpha: 0-transparent; 255-opaque
        w.setStyleSheet('color: rgba(0, 0, 0, 196); '
                        'background-color: rgba(0, 0, 0, 0);'
                        'padding: %d' % _LABEL_PADDING)
        w.setAlignment(QtCore.Qt.AlignRight)
        w.setFont(_LABEL_FONT)
        w.setGeometry(0, 0, 100, _LABEL_HEIGHT)
        self._updatePosLabelText((0, 0))
        self._updatePosLabelPosition()

    def zoomIn(self):  # noqa: N802
        """
        Increase the zoom.
        """
        self.scale(self._zoomfactor, self._zoomfactor)

    def zoomOut(self):  # noqa: N802
        self.scale(1 / self._zoomfactor, 1 / self._zoomfactor)

    def notifyPosChanged(self, turtle, pos):  # noqa: N802
        if turtle is self._scene.mainTurtle():
            self._updatePosLabelText(pos)

    def resizeEvent(self, event):  # noqa: N802
        self._updatePosLabelPosition()
        super().resizeEvent(event)

    def _updatePosLabelPosition(self):  # noqa: N802
        size = self.viewport().size()
        w = self._posLabel
        margin = 3
        w.move(size.width() - w.width() - margin,
               size.height() - w.height() - margin)

    def _updatePosLabelText(self, pos):  # noqa: N802
        s = "x=%s, y=%s" % (round(pos[0]), round(pos[1]))
        self._posLabel.setText(s)

    def saveImage(self, fname):  # noqa: N802
        """
        Saves current viewport as a png file of the given fname.
        """

        rect = self.viewport()
        rgb = QtGui.QImage.Format_RGB32
        image = QtGui.QImage(rect.width(), rect.height(), rgb)
        image.fill(QtGui.QColor(255, 255, 255))
        painter = QtGui.QPainter(image)
        self.render(painter)
        if not image.save(fname):
            raise ValueError('could not save image %s' % fname)
        del painter



"""
This module is not imported by the user, but rather it generates translation
strings for transpyler functions.
"""

import types
from collections import defaultdict
from functools import singledispatch

from .extract import extract_translations
from .gettext import gettext_for
from ..utils.decorators import synonyms as _synonyms
from ..utils.namespaces import collect_synonyms


def translate_namespace(ns, lang, synonyms=True, add_unaccented=True):
    """
    Return a dict with the given namespace translated according to the
    requested ``lang``.
    """

    _ = gettext_for(lang).gettext
    translation_data = defaultdict(dict)
    translation_strings = extract_translations(ns)
    for name, value in translation_strings.items():
        # Translate methods in a class
        if ':' in name:
            classname, method = name.split(':')
            name, sep, post = method.partition('.')
            class_dict = translation_data[classname]
            method_dict = class_dict.setdefault(name, {})
            method_dict[post] = _(value)

        # Translate a function or regular object
        else:
            name, sep, post = name.partition('.')
            translation_data[name][post] = _(value)

    # Create namespace
    namespace = dict(ns)
    for name, data in translation_data.items():
        if name not in ns:
            continue
        obj = ns[name]
        translated_name = _(name).partition('\n')[0].strip()
        namespace[translated_name] = apply_translations(obj, data)

    # Add synonyms and unaccented versions
    if synonyms:
        extra = collect_synonyms(namespace, add_unaccented=add_unaccented)
        namespace.update(extra)
    return namespace


@singledispatch
def apply_translations(obj, data):
    """
    Translate object by applying translation data to object.
    """
    return obj


@apply_translations.register(types.FunctionType)
def apply_translations_function(func, data: dict):
    """
    Return a translated version of func.
    """

    names = data['name'].strip().splitlines()
    name, *synonyms = names
    varnames = tuple(map(str.strip, data.get('args', '').split(',')))
    varnames = tuple(x for x in varnames if x)
    assert len(varnames) == len(func.__code__.co_varnames), \
        '%s: size of argument names list changed during translation (from %s ' \
        'to %s)' % (name, func.__code__.co_varnames, varnames)

    #
    # Copy the code object and change the varnames parameter
    # (Not for the faint of heart)
    #
    # code(argcount, kwonlyargcount, nlocals, stacksize, flags, codestring,
    #  constants, names, varnames, filename, name, firstlineno,
    #  lnotab[, freevars[, cellvars]])
    #
    code = func.__code__
    new_code = types.CodeType(
        code.co_argcount, code.co_kwonlyargcount, code.co_nlocals,
        code.co_stacksize, code.co_flags, code.co_code, code.co_consts,
        code.co_names, varnames, code.co_filename, code.co_name,
        code.co_firstlineno, code.co_lnotab, code.co_freevars,
        code.co_cellvars
    )

    # Create a function copy
    translated = types.FunctionType(
        new_code,
        func.__globals__,
        name=name,
        argdefs=func.__defaults__,
        closure=func.__closure__
    )
    translated.__dict__.update(func.__dict__)
    translated.__doc__ = data.get('doc', '')
    translated.__kwdefaults__ = func.__kwdefaults__
    return _synonyms(*synonyms)(translated)


@apply_translations.register(type)
def apply_translations_type(cls: type, data: dict):  # noqa: F811
    """
    Extend class with translated method.

    Changes the name and other parameters *inplace*.
    """

    return cls


def translator_factory(lang):
    """
    Return a translator function that receives a string and return a translated
    version using gettext.
    """
    gettext = gettext_for(lang)
    return gettext.gettext



#
# Random utility functions
#
import sys


def with_transpyler_attr(cls, transpiler):
    """
    Return a subclass with the transpyler attribute defined.
    """

    return type(cls.__name__, (cls,), {'transpyler': transpiler})


def has_qt():
    """
    Returns True if PyQt5 is installed.
    """
    try:
        import PyQt5  # noqa: F401
        return True
    except ImportError:
        return False


def clear_argv(argv=sys.argv):
    """
    Remove extra arguments from sys.argv.
    """

    keep = [argv.pop(0)]
    if keep[0] == sys.executable:
        keep.append(argv.pop(0))
    argv[:] = keep



import functools
from gettext import gettext as _

from unidecode import unidecode


def synonyms(*args):
    """
    Decorator that marks synonyms/aliases of a function.

    Examples:

        .. code-block:: python

            @synonyms('fd', 'fwd')
            def forward(x):
                print('go forward')
    """

    fmt_args = [':func:`%s`' % x for x in args]

    def decorator(func):
        synonyms = getattr(func, '__synonyms__', ())
        func.__synonyms__ = tuple(synonyms) + args

        if hasattr(func, '__doc__'):
            data = ', '.join(fmt_args[:])
            func.__doc__ = \
                (func.__doc__ or '') + \
                ('\n\n'
                 'Notes:\n'
                 '    %s: %s') % (_('Synonyms'), data)
        return func

    return decorator


def normalize_accented_keywords(func):
    """
    Decorator that remove accents from keyword names and pass unaccented
    versions to the implementation function.
    """

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        decode = unidecode
        kwargs = {decode(k): v for (k, v) in kwargs.items()}
        return func(*args, **kwargs)

    return decorated


#
# Make objects that execute code with side effects with a simple call of repr()
# This is useful in the terminal to implement "commands" such as exit
#
class PrettyCallable:
    """
    Callable that whose repr() is a given message.
    """

    def __init__(self, func, name=None, doc=None, str=None,
                 autoexec=False, autoexec_message=None):
        self.__func = func
        self.__autoexec = autoexec
        self.__autoexec_message = autoexec_message
        self.__repr = str or 'please call %s()' % self.__name__

    def __call__(self, *args, **kwargs):
        return self.__func(*args, **kwargs)

    def __repr__(self):
        if self.__autoexec:
            self.__func()
            return self.__autoexec_message or ''
        return self.__repr

    def __getattr__(self, attr):
        return getattr(self.__func, attr)


def pretty_callable(str=None, **kwargs):
    """
    Decorate function to be a pretty callable.

    Example:
        >>> @pretty_callable('call exit() to finish interactive shell')
        ... def exit():
        ...     raise SystemExit
    """

    def decorator(func):
        return PrettyCallable(func, str=str, **kwargs)

    return decorator



# flake8: noqa
from .decorators import synonyms, normalize_accented_keywords, pretty_callable
from .string import keep_spaces, humanize_name, unhumanize_name, \
    normalize_docstring, split_docstring
from .utils import with_transpyler_attr, clear_argv, has_qt
from .namespaces import full_class_name, collect_synonyms, extract_namespace



from .utils.utils import has_qt

def run(transpyler):
    """
    Starts the default main application.
    """
    import click

    @click.command()
    @click.option('--cli', '-c', is_flag=True, default=False,
                  help='start gui-less console.')
    @click.option('--console', is_flag=True, default=False,
                  help='start a simple gui-less console.')
    @click.option('--notebook/--no-notebook', '-n', default=False,
                  help='starts notebook server.')
    def main(cli, notebook, console):
        command = None

        if cli:
            command = lambda: start_console(transpyler, console='auto')
        elif console:
            command = lambda: start_console(transpyler, console='console')
        elif notebook:
            command = lambda: start_notebook(transpyler)
        elif has_qt():
            command = lambda: start_qturtle(transpyler)
        else:
            click.echo('Could not start GUI. Do you have Qt installed?',
                           err=True)
            command = lambda: start_console(transpyler, console='nogui')

        return command()

    return main()

def start_console(transpyler, console='auto', turtle='auto'):
    """
    Starts a regular python console with the current transpyler.

    Args:
        console:
            Can be one of 'jupyter', 'console', 'qtconsole', 'auto'. This
            chooses the default console application. The default behavior
            (auto) is to try jupyter and fallback to console if it is
            not available.
        turtle:
            Sets the turtle backend. It can be either 'qt', 'tk', 'none' or
            'auto'. The defaut strategy (auto) is to try the qt first and
            fallback to tk.
    """

    # Select the console application
    if console == 'auto':
        try:
            import IPython  # noqa: F401
        except ImportError:
            console = 'console'
        else:
            console = 'jupyter'

    if console == 'qtconsole':
        from .jupyter import start_jupyter
        start_jupyter(transpyler=transpyler, gui=True)

    elif console == 'jupyter':
        from .jupyter import start_jupyter
        start_jupyter(transpyler=transpyler, gui=False)

    elif console == 'console':
        from .console import start_console
        start_console(transpyler=transpyler)

def start_notebook(transpyler):
    """
    Starts a jupyter notebook with the current transpyler.
    """

    from .jupyter import start_notebook
    start_notebook(transpyler)

def start_qturtle(transpyler):
    """
    Starts a QTurtle application with the current transpyler.
    """

    if not has_qt():
        raise SystemExit('PyQt5 is necessary to run the turtle '
                         'application.')

    from qturtle.mainwindow import start_application
    start_application(transpyler)



import types
from collections import OrderedDict
from functools import singledispatch

from ..utils import normalize_docstring


#
# Extract translations: these functions generate a .pot file from a python
# module
#
def extract_translations(namespace):
    """
    Extract all translation strings from namespace.

    Return a dictionary mapping ids to their corresponding values. The values
    would typically be inserted in a POT file. Translators would then generate
    PO files for each desired language.

    >>> extract_translations(ns)                                # doctest: +SKIP
    {
        'cos.name': 'cos',
        'cos.args': 'x',
        'cos.doc': 'Returns the cosine of a number',
        ...
    }

    Notes:
        If the namespace have a 'TRANSLATIONS' key mapping to a dictionary,
        it is used to update the result. The contents of TRANSLATIONS override
        any translation string computed automatically.
    """
    result = OrderedDict()
    extra_translations = namespace.get('TRANSLATIONS', {})

    for k, v in namespace.items():
        if k.startswith('_') or k == 'TRANSLATIONS':
            continue

        translations = extract_translation(v)
        if translations:
            for name, st in translations.items():
                name = name if name.startswith(':') else '.' + name
                result[k + name] = st
        else:
            result[k] = k

    result.update(extra_translations)
    return result


@singledispatch
def extract_translation(obj):
    """
    Returns a dictionary mapping paths inside a single object to translatable
    strings. If there is nothing to translate, it returns an empty dict.
    """

    # Objects that define __name__, __doc__ and __synonyms__
    result = OrderedDict()

    if hasattr(obj, '__name__'):
        result['name'] = obj.__name__
        for syn in getattr(obj, '__synonyms__', ()):
            result['name'] += '\n' + syn

    if hasattr(obj, '__doc__') and obj.__doc__:
        result['doc'] = normalize_docstring(obj.__doc__)

    return result


@extract_translation.register(types.FunctionType)
def extract_translation_function(obj):
    result = extract_translation.dispatch(object)(obj)
    vars = obj.__code__.co_varnames
    if vars:
        result['args'] = ', '.join(vars)
    return result


@extract_translation.register(type)
def extract_translation_type(obj):
    result = extract_translation.dispatch(object)(obj)

    for attr in dir(obj):
        if attr.startswith('_'):
            continue
        method = getattr(obj, attr, None)

        if isinstance(method, types.FunctionType):
            trans = extract_translation(method)
            for k, v in trans.items():
                result[':%s.%s' % (attr, k)] = v
        else:
            result[':%s' % attr] = attr

    return result



from .gettext import gettext_for, create_pot_file, L10N_PATH, gettext, set_language
from .extract import extract_translations, extract_translation
from .translate import translate_namespace, apply_translations, translator_factory



import os

import polib
from textblob import TextBlob as Blob
from textblob.exceptions import NotTranslated

from .gettext import L10N_PATH, PO_HEADER
from ..utils.string import split_docstring


def google_translate(lang, verbose=True, prompt=True):
    """
    Fill translations for the given PO file using google translate.
    """

    # Utility functions
    echo = print if verbose or prompt else lambda *args, **kwargs: None
    ask = _ask(verbose, prompt)
    showline = _showline(verbose, prompt)

    # Prepare
    popath = os.path.join(L10N_PATH, lang + '.po')
    if os.path.exists(popath):
        echo('Loading file: %r' % popath, '\n\n')
    else:
        echo('Creating file: %r' % popath, '\n\n')
        potpath = os.path.join(L10N_PATH, 'transpyler.pot')
        save_po_file(potpath, popath, lang)
    pofile = polib.pofile(popath)

    translator = GoogleTranslator(lang)

    # Mainloop: handle all missing translations and pass them to google
    # translate
    for entry in pofile:
        if entry.msgstr:
            continue

        echo('\n'.join('#. ' + line for line in entry.comment.splitlines()))
        echo(entry.msgid)
        if ask():
            id = entry.comment.rpartition('.')[-1]
            entry.msgstr = translator(entry.msgid, id)
            entry.flags.append('fuzzy')

            # Show message
            showline()
            echo('\nTRANSLATION\n')
            echo(entry.msgstr)
            showline()

            # Save
            pofile.save()


def save_po_file(potpath, popath, lang):
    """
    Saves .po file using data from potfile and language.
    """

    with open(popath, 'w', encoding='utf-8') as F:
        F.write(PO_HEADER % lang.replace('-', '_'))

        with open(potpath, encoding='utf8') as Fpot:
            save = False
            for line in Fpot.readlines():
                if save:
                    F.write(line)
                elif line.isspace():
                    save = True


def _showline(verbose, prompt):
    return lambda: \
        print('\n' + '-' * 40) if verbose or prompt else lambda: None


def _ask(verbose, prompt):
    return lambda: (
        input('\nTranslate? [Y/n]').lower() in ('y', '')
        if prompt else lambda: True
    )


class GoogleTranslator:
    """
    A namespace with utility functions used to implement translations with
    google translate.
    """

    def __init__(self, lang):
        self.lang = lang
        self.google_lang = lang.replace('_', '-')

    def __call__(self, text, id):
        map = {
            'doc': self.translate_doc,
            'name': self.translate_name,
            'args': self.translate_args,
        }
        return map.get(id, self.google_translate)(text)

    def google_translate(self, x):
        if not x:
            return ''
        try:
            result = '%s' % Blob(x).translate('en', self.google_lang)
        except NotTranslated:
            result = x

        if x[0].isupper():
            return result
        else:
            return result[0].lower() + result[1:]

    def translate_doc(self, text):
        sections = []
        for section in split_docstring(text):
            translated = self.google_translate(section)
            sections.append(translated)
        return '\n\n'.join(map(lambda x: '%s' % x, sections))

    def translate_name(self, name):
        translated = self.google_translate(name.replace('_', ' '))
        if name.islower():
            translated = translated.lower()
        elif name[0].islower():
            translated = translated[0].lower() + translated[1:]
        return translated.replace(' ', '_').replace('-', '_')

    def translate_args(self, text):
        return ', '.join(map(self.translate_name, text.split(',')))



import builtins as _builtins
import time as _time


#
# Standard Python functions
#
def range(*args):
    """
    range(stop) -> range object
    range(start, stop[, step]) -> range object

    Return an object that produces a sequence of integers from start (inclusive)
    to stop (exclusive) by step.  range(i, j) produces i, i+1, i+2, ..., j-1.
    start defaults to 0, and stop is omitted!  range(4) produces 0, 1, 2, 3.
    These are exactly the valid indices for a list of 4 elements.
    When step is given, it specifies the increment (or decrement).
    """
    return _range(*args)


#
# Time control
#
def sleep(seconds):
    """
    Delay execution for a given number of seconds.

    The argument may be a floating point number for subsecond precision.
    """
    return _sleep(seconds)


def exit(msg=0):
    """
    Finish program execution and provides an exit code.

    If msg is not given, the program finishes in a successful state. An
    exit code different from zero or a string tells that there were an error
    during execution.
    """
    return _exit(msg)


#
# Conversions/representations of numerical types
#
def integer(x):
    "Converts argument to an integer."
    return _integer(x)


def real(x):
    "Convert argument to a real number (number with decimal places)."
    return _real(x)


def complex(x):
    "Convert argument to a complex number."
    return _complex(x)


def boolean(x):
    """Convert argument to a boolean value.

    Zero and empty sequences are converted to False. All other values are True.
    """
    return _boolean(x)


def binary(x):
    "Convert integer to a string with its binary representation."
    return _binary(x)


def octal(x):
    "Convert integer to a string with its octal representation."
    return _octal(x)


def hexadecimal(x):
    "Convert integer to a string with its hexadecimal representation"
    return _hexadecimal(x)


def character(x):
    """
    Convert number to its corresponding character in the ASCII/Unicode tables.
    """
    return _character(x)


# These functions can be replaced/mocked and have a proper behavior in a
# graphical environment or in tests.
_range = _builtins.range
_sleep = _time.sleep
_integer = _builtins.int
_real = _builtins.float
_complex = _builtins.complex
_string = _builtins.str
_boolean = _builtins.bool
_binary = _builtins.bin
_octal = _builtins.oct
_hexadecimal = _builtins.hex
_character = _builtins.chr


def _exit(msg):
    raise SystemExit(msg)



import collections

from .utils import vecargsmethod
from ..math import Vec


class TurtleGroup(collections.MutableSequence):
    """
    A synchronized group of turtles.

    Functions are broadcasted to all turtles belonging to the group.
    """

    _state_class = None
    __slots__ = ('_data',)

    def __init__(self):
        self._data = []

    def __repr__(self):
        return 'TurtleGroup(%r)' % self._data

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __getitem__(self, i):
        return self._data[i]

    def __delitem__(self, i):
        return self._data.__delitem__(i)

    def __setitem__(self, i, v):
        return self._data.__setitem__(i, v)

    def insert(self, i, v):
        return self.insert(i, v)

    @vecargsmethod
    def setpos(self, value):
        pos = Vec(*value)
        for t in self:
            t.setpos(pos)

    def setheading(self, value):
        for t in self:
            t.setheading(value)

    def setwidth(self, value):
        for t in self:
            t.setwidth(value)

    def setcolor(self, value):
        for t in self:
            t.setcolor(value)

    def setfillcolor(self, value):
        for t in self:
            t.setfillcolor(value)

    def setavatar(self, value):
        for t in self:
            t.setavatar(value)

    def penup(self):
        """
        Raises the turtle pen so it stops drawing.

        Aliases: penup | pu
        """
        for t in self:
            t.penup()

    def pendown(self):
        """
        Lower the turtle pen so it can draw in the screen.

        Aliases: pendown | pd
        """
        for t in self:
            t.pendown()

    def isdown(self):
        """
        Return True if all turtles are drawing.
        """
        return all(t.isdown() for t in self)

    def isvisible(self):
        """
        Return True if all turtles are visible.
        """
        return all(t.isvisible() for t in self)

    def arehidden(self):
        """
        Return True if all turtles are not visible.
        """
        return all(t.ishidden() for t in self)

    def hide(self):
        """
        Hide all turtles.
        """
        for t in self:
            t.hide()

    def show(self):
        """
        Shows all hidden turtle.
        """
        for t in self:
            t.show()

    @vecargsmethod
    def goto(self, pos):
        """
        Goes to the given position.

        If the pen is down, it draws a line.
        """
        for t in self:
            t.goto(pos)

    @vecargsmethod
    def jump(self, pos):
        """
        Relative movement by the desired position. It *never* draw as line even
        if the pen is down.
        """
        for t in self:
            t.jump(pos)

    def forward(self, step):
        """
        Move the turtle forward by the given step size (in pixels).

        Aliases: forward | fd
        """
        for t in self:
            t.forward(step)

    def backward(self, step):
        """
        Move the turtle backward by the given step size (in pixels).

        Aliases: backward | back | bk
        """
        for t in self:
            t.backward(step)

    def left(self, angle):
        """
        Rotate the turtle counter-clockwise by the given angle.

        Aliases: left | lt

        Negative angles produces clockwise rotation.
        """
        for t in self:
            t.left(angle)

    def right(self, angle):
        """
        Rotate the turtle clockwise by the given angle.

        Aliases: right | rt

        Negative angles produces counter-clockwise rotation. Return final
        heading.
        """
        for t in self:
            t.right(angle)

    def clean(self):
        """
        Clear all drawings made by turtles.
        """
        for t in self:
            t.clean()

    def reset(self):
        """
        Clear all drawings and reset turtles to initial position.
        """
        for t in self:
            t.reset()

    # Aliases
    fd = forward
    bk = back = backward
    lt = left
    rt = right
    pu = penup
    pd = pendown



from .utils import turtle_property
from .namespace import TurtleNamespace
from . import Turtle as BaseTurtle, PropertyState


class TkTurtleState(PropertyState):
    """
    A turtle state class for Python's builtin turtle module.
    """

    valid_avatars = [
        'classic', 'arrow', 'turtle', 'circle', 'square', 'triangle'
    ]

    pos = _pos = turtle_property('pos', 'setpos')
    heading = _heading = turtle_property('heading', 'setheading')
    color = _color = turtle_property('color')
    fillcolor = _fillcolor = turtle_property('fillcolor')
    width = _width = turtle_property('width')
    avatar = _avatar = turtle_property('shape')

    def getdrawing(self):
        return self.turtle.isdown()

    def setdrawing(self, value):
        self.turtle.pd() if value else self.turtle.pu()

    def gethidden(self):
        return self.turtle.isvisible()

    def sethidden(self, value):
        self.turtle.hideturtle() if value else self.turtle.showturtle()

    def __init__(self, *args, **kwargs):
        from turtle import Turtle

        self.turtle = Turtle()
        super().__init__(*args, **kwargs)

    def step(self, step):
        self.turtle.fd(step)

    def rotate(self, angle):
        self.turtle.lt(angle)

    def move(self, pos):
        self.turtle.goto(*pos)


# The tk turtle class
class Turtle(BaseTurtle):
    """
    Represents a turtle.
    """

    _state_factory = TkTurtleState


def make_turtle_namespace():
    """
    Returns a dictionary with the namespace of turtle functions.
    """

    return dict(TurtleNamespace(Turtle))


def start_console(transpyler=None):
    """
    Starts a console with turtle functions based on the Tk module.
    """

    from transpyler.console import start_console
    start_console(make_turtle_namespace(), transpyler=transpyler)


# We load a global namespace if called with the python -m flag.
if __name__ == '__main__':
    start_console()



import re

SPACE_RE = re.compile('^[ ]*')
HEADING_SPACE = re.compile(r'^\s*')
ENDING_SPACE = re.compile(r'\s*$')


def humanize_name(name):
    """
    Convert a variable or class name to a version that a human can understand.

    Examples:
        >>> humanize_name('SomeName')
        'Some Name'
    """

    name = name.replace('_', ' ')
    letters = []
    for c in name:
        if c.isupper() and letters and letters[-1] != '_':
            letters.append(' ' + c)
        else:
            letters.append(c)
    return ''.join(letters)


def unhumanize_name(name):
    """
    Revert the actions of humanize_name.
    """

    pieces = name.split()
    while len(pieces) > 1:
        if pieces[1][0].isupper():
            pieces[0] += pieces.pop(1)
        else:
            pieces[0] += '_' + untitle(pieces.pop(1))
    return pieces[0]


def untitle(st):
    """
    Revert the effect of str.title()
    """
    if st and st[0].isupper():
        return st[0].lower() + st[1:]
    return st


def keep_spaces(result, src):
    """
    Keep the same number of heading and trailing whitespace in result as
    compared to src.

    Example:
        >>> keep_spaces(' foo', 'bar\n\n')
        'foo\n\n'
    """

    if not src:
        return result.strip()

    # Find head and tail
    head = tail = ''
    if src[0].isspace():
        head = HEADING_SPACE.search(src).group()
    if src[-1].isspace():
        tail = ENDING_SPACE.search(src).group()
    return head + result.strip() + tail


def normalize_docstring(doc):
    """
    Normalize docstring.
    """

    doc = doc.strip().replace('\n\t', '\n    ')
    if not doc:
        return doc

    lines = doc.splitlines()
    indents = [SPACE_RE.match(line).span()[1] for line in lines]
    zipped = list(zip(indents, lines))

    if indents[0] == 0:
        zipped = zipped[1:]
    ind = min(zipped, default=(0, ''), key=lambda x: (x[0] if x[1] else 80))[0]

    for i, line in enumerate(lines):
        if i == 0 and indents[0] == 0:
            continue
        lines[i] = '' if not line.strip() else line[ind:]

    return '\n'.join(lines)


def split_docstring(text):
    """
    Split docstring in several sections.
    """

    parts = ['']
    text = normalize_docstring(text)
    parts[0], *lines = text.splitlines()

    for line in lines:
        if not line:
            parts.append('')
        elif line.startswith(' '):
            parts[-1] += '\n' + line
        elif parts[-1] == '':
            parts[-1] += line
        else:
            parts[-1] += ' ' + line

    return parts



from multiprocessing import Queue

from colortools import Color

from .utils import getsetter, ipc_property
from ..math import vec, cos, sin, tan


class TurtleState:
    """
    An object that stores the turtle state. Users interact with Turtle
    instances that manipulates a state object. The state can be local and
    apply changes immediately or it can be remote and echo to a state stored
    in a different process/machine/thread.
    """

    valid_avatars = ['default']

    @property
    def heading_direction(self):
        return vec(self._cos(self.heading), self._sin(self.heading))

    def __init__(self, pos=None, heading=0.0, drawing=True,
                 color='black', fillcolor='black', width=1, hidden=False,
                 avatar=None, group=None, id=None):
        self.pos = self.startpos = self._vec(pos or (0, 0))
        self.heading = self.startheading = heading
        self.drawing = self.startdrawing = drawing
        self.color = color
        self.fillcolor = fillcolor
        self.width = width
        self.hidden = hidden
        self.avatar = self.valid_avatars[0] if avatar is None else avatar
        self.lines = []
        self.group = group
        self.id = id

    def validate_avatar(self, avatar):
        """
        Raises a ValueError if avatar is not a valid avatar name.
        """
        if avatar is not None and avatar not in self.valid_avatars:
            raise ValueError('invalid avatar name: %s')

    def rotate(self, angle):
        """
        Rotate cursor heading in the counter clockwise direction.
        """
        self.heading += angle

    def move(self, pos):
        """
        Move turtle possibly drawing a line.
        """
        oldpos = self.pos
        self.pos = pos
        if self.drawing:
            self.draw_line(oldpos, pos)

    def step(self, step):
        """
        Move forwards (or backwards if step is negative).
        """
        self.move(self.pos + self.heading_direction * step)

    def clean(self):
        """
        Remove all drawn lines.
        """
        self.lines.clear()

    def reset(self):
        """
        Reset turtle to initial state.
        """

        self.clean()
        self.pos = self.startpos
        self.heading = self.startheading
        self.drawing = self.startdrawing

    def draw_line(self, v1, v2):
        """
        Draws line from v1 to v2.
        """
        raise NotImplementedError

    def register(self, group):
        """
        Register on a group.
        """
        self.group = group

    def recv(self):
        """
        Receives messages from a connection.
        """
        raise NotImplementedError

    def handle(self, msg):
        """
        Handle message and redirect to the proper method.
        """

        action, id, *args = msg
        if id != self.id:
            raise ValueError('invalid turtle id (%s)' % msg)

        if action == 'get':
            return ['get', id, getattr(self, args[0])]
        elif action == 'set':
            setattr(self, *args)
            return ['set', id, args[0]]
        elif action == 'rotate':
            self.rotate(args[0])
            return ['rotate', id]
        elif action == 'move':
            self.move(args[0])
            return ['move', id]
        elif action == 'step':
            self.step(args[0])
            return ['step', id]
        elif action == 'clean':
            self.clean()
            return ['clean', id]
        else:
            raise ValueError('invalid action: %r' % action)

    # Mathematical functions. We define them here so they can be overridable in
    # subclasses.
    _cos = staticmethod(cos)
    _sin = staticmethod(sin)
    _tan = staticmethod(tan)
    _vec = staticmethod(vec)


class PropertyState(TurtleState):
    """
    A Turtle state in which all attributes are properties that expose a
    getter/setter interface.

    This state class is used by the Tk implementation.
    """

    # Position
    getpos = (lambda self: self._pos)
    setpos = (lambda self, v: setattr(self, '_pos', vec(*v)))
    pos = getsetter('pos')

    # Heading
    getheading = (lambda self: self._heading)
    setheading = (lambda self, v: setattr(self, '_heading', v))
    heading = getsetter('heading')

    # Drawing
    getdrawing = (lambda self: self._drawing)
    setdrawing = (lambda self, v: setattr(self, '_drawing', v))
    drawing = getsetter('drawing')

    # Color
    getcolor = (lambda self: self._color)
    setcolor = (lambda self, v: setattr(self, '_color', v))
    color = getsetter('color')

    # Fillcolor
    getfillcolor = (lambda self: self._fillcolor)
    setfillcolor = (lambda self, v: setattr(self, '_fillcolor', v))
    fillcolor = getsetter('fillcolor')

    # Width
    getwidth = (lambda self: self._width)
    setwidth = (lambda self, v: setattr(self, '_width', v))
    width = getsetter('width')

    # Hidden
    gethidden = (lambda self: self._hidden)
    sethidden = (lambda self, v: setattr(self, '_hidden', v))
    hidden = getsetter('hidden')

    # Avatar
    getavatar = lambda self: self._avatar

    def setavatar(self, v):
        self.validate_avatar(v)
        self._avatar = v

    avatar = getsetter('avatar')


class RemoteState(TurtleState):
    """
    An abstract state class that stores information remotely and pipes all
    requests to a `.send(msg)` method. Subclass must implement the .send()
    method accordingly.

    The message is an S-expression:

    ['get', id, attr]:
        Get an state attribute
    ['set', id, attr, value]:
        Set an state attribute to a given value
    ['step', id, length]:
        Steps forward for the given length.
    ['rotate', id, angle]:
        Rotates by the given angle.
    ['move', id, pos]:
        Moves turtle to the given position.

    The actual state info is stored on another thread or process.
    """

    pos = ipc_property('pos', vec, tuple)
    heading = ipc_property('heading')
    drawing = ipc_property('drawing')
    color = ipc_property('color', Color, tuple)
    fillcolor = ipc_property('fillcolor', Color, tuple)
    width = ipc_property('width')
    hidden = ipc_property('hidden')
    avatar = ipc_property('avatar')
    id = None

    def __init__(self, **kwargs):
        msg = self.send(['newturtle', kwargs])
        if msg[0] != 'newturtle':
            raise ValueError('invalid reply: %s' % msg)
        self.id = kwargs.pop('id', msg[1])
        if not self.id:
            raise ValueError('invalid remote id: %s' % self.id)

    def rotate(self, angle):
        self.send(['rotate', self.id, angle])

    def move(self, pos):
        self.send(['move', self.id, tuple(pos)])

    def step(self, step):
        self.send(['step', self.id, step])

    def getvalue(self, attr):
        return self.send(['get', self.id, attr])[2]

    def setvalue(self, attr, value):
        return self.send(['set', self.id, attr, value])

    def clean(self):
        self.send(['clear', self.id])

    def send(self, msg):
        """
        Implemented in the client process. Just sends messages through a
        connection.
        """
        raise NotImplementedError('must implement the .send(msg) method')


class MailboxState(RemoteState):
    """
    A remote state that uses a multiprocessing.Queue to communicate between
    processes.

    The communication uses an inbox/outbox model. The client puts a message in
    the outbox and the server responds with a message in the inbox.
    """

    def inbox_factory(self):
        return Queue()

    def outbox_factory(self):
        return Queue()

    timeout = 1.0

    def __init__(self, **kwargs):
        # Initialize inbox
        try:
            self.inbox = kwargs['inbox']
        except KeyError:
            if not hasattr(self, 'inbox'):
                self.inbox = self.inbox_factory()

        # Initialize outbox
        try:
            self.outbox = kwargs['outbox']
        except KeyError:
            if not hasattr(self, 'outbox'):
                self.outbox = self.outbox_factory()

        super().__init__(**kwargs)

    def send(self, msg):
        self.put_message(msg)
        return self.get_message()

    def put_message(self, msg):
        """
        Put outgoing message on the outbox
        """
        self.outbox.put(msg)

    def get_message(self):
        """
        Get incoming message from the inbox.
        """
        return self.inbox.get(timeout=self.timeout)

    def draw_line(self, v1, v2):
        pass  # no-op: we broadcast and let the server handle it


class MirrorState(RemoteState):
    """
    A remote turtle that stores a copy of state in itself.

    This state is used by the QTurtle application in the kernel process.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.local = TurtleState(**kwargs)
        self.local.draw_line = lambda v1, v2: None
        self.id = self.local.id
        self.lines = []

    def getvalue(self, attr):
        return getattr(self.local, attr)

    def setvalue(self, attr, value):
        setattr(self.local, attr, value)
        super().setvalue(attr, value)

    def rotate(self, angle):
        local = self.local
        local.rotate(angle)
        super().rotate(angle)
        self.setvalue('heading', local.heading)

    def move(self, pos):
        local = self.local
        local.move(pos)
        super().move(pos)
        self.setvalue('pos', local.pos)

    def step(self, step):
        local = self.local
        local.step(step)
        super().step(step)
        self.setvalue('pos', local.pos)

    def clean(self):
        self.local.clean()
        super().clean()

    def draw_line(self, v1, v2):
        pass



import operator as op

import colortools
from PyQt5.QtCore import QPointF, QObject
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QGraphicsLineItem

from transpyler.math import Vec
from transpyler.turtle.utils import webcolor


def to_qcolor(value):
    """
    Convert a color, color tuple or color string to a QColor object.
    """
    if isinstance(value, str):
        value = webcolor(value)
    color = colortools.Color(value)
    return QColor(*color.rgba)


def from_qcolor(color):
    """
    Converts QColor to Color object.
    """
    return colortools.Color(*color.rgba())


def to_qvector(value):
    """
    Converts a vector-like object to QPointF.
    """
    return QPointF(*value)


def from_qvector(value):
    """
    Converts a QPointF to a Vec.
    """
    return Vec(value.x(), value.y())


class QLine(QGraphicsLineItem, QObject):
    """
    A line/QObject class.
    """

    def __init__(self, v0, v1, pen):
        x0, y0 = v0
        x1, y1 = v1
        super().__init__(x0, y0, x1, y1)
        self.setPen(pen)


def qtproperty(name, unwrap=lambda x: x, wrap=lambda x: x):
    """
    Expose a property of an attribute that respects Qt's get/setter
    conventions.

    If transform is defined, it is applied to the value passed to the setter
    method.

    Example::

        class MyObj:
            title = qtproperty('_widget.title')
    """

    prop, action = name.split('.')
    set_name = '%s.set%s' % (prop, action.title())
    getter = op.attrgetter(name)
    setter = op.attrgetter(set_name)
    return property(
        lambda x: unwrap(getter(x)()),
        lambda x, v: setter(x)(wrap(v)),
    )



import importlib
from ast import PyCF_ONLY_AST

from IPython.core.compilerop import CachingCompiler
from IPython.utils import py3compat
from ipykernel.ipkernel import IPythonKernel
from lazyutils import lazy
from traitlets import Type

from .shell import TranspylerShell
from transpyler.utils import with_transpyler_attr


class TranspylerKernel(IPythonKernel):
    """
    A meta kernel based backend to use Transpyled languages in Jupyter/iPython.
    """

    transpyler = None
    implementation = lazy(lambda self: 'i' + self.transpyler.name)
    implementation_version = lazy(lambda self: self.transpyler.version)
    language = lazy(lambda self: self.transpyler.name)
    language_version = lazy(lambda self: self.transpyler.language_version)
    banner = lazy(lambda self: self.transpyler.console_banner())
    language_info = lazy(lambda self: self.transpyler.info.get_language_info())
    shell_class = Type(TranspylerShell)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.transpyler is None:
            raise ValueError('transpyler was not defined')
        monkey_patch(self.transpyler)
        self.transpyler.init()

    def do_execute(self, code, *args, **kwargs):
        code = self.transpyler.transpile(code)
        return super().do_execute(code, *args, **kwargs)

    def do_is_complete(self, code):
        return super().do_is_complete(self.transpyler.transpile(code))


def start_kernel(transpyler):
    """
    Start Jupyter kernel.
    """

    from ipykernel.kernelapp import IPKernelApp

    kernel_class = with_transpyler_attr(TranspylerKernel, transpyler)
    IPKernelApp.launch_instance(kernel_class=kernel_class)


# We mokey-patch a few modules and functions to replace Python's compile
# function with pytuga's version. This is potentially fragile, but it seems to
# work well so far.
def monkey_patch(transpyler):
    def ast_parse(self, source, filename='<unknown>', symbol='exec'):
        flags = self.flags | PyCF_ONLY_AST
        return transpyler.compile(source, filename, symbol, flags, 1)

    CachingCompiler.ast_parse = ast_parse
    py3compat.compile = transpyler.compile


# Find correct transpyler instance from argv and execute start_kernel
if __name__ == '__main__':
    import sys

    idx = sys.argv.index('--type')
    sys.argv.pop(idx)
    transpyler_class_name = sys.argv.pop(idx)
    path, _, name = transpyler_class_name.rpartition('.')
    transpyler_class = getattr(importlib.import_module(path), name)
    start_kernel(transpyler_class())



from collections import deque
from multiprocessing import Queue

from PyQt5 import QtWidgets, QtCore

from .state import QGraphicsSceneGroup


class TurtleScene(QtWidgets.QGraphicsScene):
    """
    The TurtleScene defines the scene in which geometric objects resides.

    It controls the Turtle object and how it draws in the screen.
    """

    clearScreenSignal = QtCore.pyqtSignal()
    restartScreenSignal = QtCore.pyqtSignal()

    def __init__(self, parent=None, fps=30, inbox=None, outbox=None):
        super().__init__(parent)
        self._fps = fps
        self._interval = 1 / fps
        self.startTimer(1000 / fps)

        # Connect signals to slots
        self.clearScreenSignal.connect(self.cleanScreen)
        self.restartScreenSignal.connect(self.restartScreen)

        # Creates mail boxes
        self._inbox = inbox = Queue() if inbox is None else inbox
        self._outbox = outbox = Queue() if outbox is None else outbox

        # Init
        self._turtles = QGraphicsSceneGroup(self, inbox=inbox, outbox=outbox)
        self._tasks = deque()
        assert self._turtles.inbox is self._inbox
        assert self._turtles.outbox is self._outbox

    def clearTurtles(self):  # noqa: N802
        self._turtles.clean()

    #
    # Events and signals
    #
    def cleanScreen(self):  # noqa: N802
        state = self.fullTurtleState()
        self.clear()
        self.setFullTurtleState(state)

    def restartScreen(self):  # noqa: N802
        self.clear()

    #
    # Task control and communication
    #
    def ping(self, receive=False):
        """
        Handles a ping message.
        """
        self._turtles.ping(receive=receive)

    def timerEvent(self, timer):  # noqa: N802
        """
        Scheduled to be executed at some given framerate.

        It process all queued animation _tasks.
        """

        inbox = self._inbox
        turtles = self._turtles
        while not inbox.empty():
            turtles.recv()

    #
    # Turtle control
    #
    def newTurtle(self, *, default=False, **kwds):  # noqa: N802
        """
        Adds new turtle to the scene and return it.
        """

        turtle = self._turtles.new_turtle(**kwds)
        self.addItem(turtle.graphic_item)
        if default:
            self.setTurtle(turtle)
        return turtle

    def turtle(self):  # noqa: N802
        """
        Return the active turtle.
        """
        return self._turtle

    def setTurtle(self, turtle):  # noqa: N802
        """
        Configures the active Turtle object.
        """
        self._turtle = turtle

    #
    # Turtle visibility
    #
    def isActiveTurtleVisible(self):  # noqa: N802
        return not self.isActiveTurtleHidden()

    def isActiveTurtleHidden(self):  # noqa: N802
        if self._turtle is None:
            return True
        return self._turtle.hidden

    def hideActiveTurtle(self):  # noqa: N802
        if self._turtle is not None:
            self._turtle.hide()

    def showActiveTurtle(self):  # noqa: N802
        if self._turtle is None:
            raise ValueError('no turtle defined')
        self._turtle.show()

    #
    # Communication with the client. The client sends messages with turtle
    # commands. The scene graph updates the turtle state and return a response.
    #
    def receive(self, msg):
        # TODO: handle this properly. Turtle should send messages that asks
        # for certain turtle movements and draws the corresponding geometric
        # figures
        pass



"""
Input and output functions
==========================

Functions for user interaction. Either reads files or typed values, or
displays values on the screen.

.. autofunction:: show
.. autofunction:: alert
.. autofunction:: show_formatted
.. autofunction:: read_text
.. autofunction:: read_number
.. autofunction:: read_file
.. autofunction:: save_in_file
.. autofunction:: pause


String processing
=================

Text processing functions.


Other
=====

Other functions that do not fall in any specific category.


Mathematical functions
======================

"""

TRANSLATIONS = {
    'exiter.name': 'quit',
    'exiter.doc': 'Type quit() to leave.',
    'cls.name': 'cls',
    'cls.doc': 'Clear screen'
}

from ._lib_std import *
from ._lib_str import *
from ._lib_io import *
from ..math import *



from colortools import Color
from lazyutils import lazy

from .utils import vecargsmethod
from ..math import Vec


class Turtle:
    """
    Creates a new Turtle.
    """

    _state_factory = None

    @lazy
    def _state(self):
        if self._state_factory is None:
            raise RuntimeError('Turtle does not define an _state_factory '
                               'attribute')
        return self._state_factory(**self._args)

    def __init__(self, pos=None, heading=0.0, *, drawing=True,
                 color='black', fillcolor='black', width=2, hidden=False):
        self._args = dict(
            pos=pos, heading=heading, drawing=drawing, color=color,
            fillcolor=fillcolor, width=width, hidden=hidden
        )

    def __repr__(self):
        return 'Turtle(%r)' % self.getavatar()

    def getpos(self):
        """
        Return a vector (x, y) with turtle's position (in pixels).
        """
        return self._state.pos

    @vecargsmethod
    def setpos(self, value):
        """
        Modifies turtle's position (in pixels)

        User can pass the x, y coordinates of the new position or a tuple of
        (x, y) values. This method never draws a line.
        """
        self._state.pos = Vec(*value)

    def getheading(self):
        """
        Return current heading of the turtle (in degrees).
        """
        return self._state.heading

    def setheading(self, value):
        """
        Sets turtle's heading (in degrees).
        """
        self._state.heading = value

    def getwidth(self):
        """
        Return the pen width (in pixels):
        """
        return self._state.width

    def setwidth(self, value):
        """
        Modifies the pen width (in pixels)
        """
        self._state.width = value

    def getcolor(self):
        """
        Return a tuple of (R, G, B) with the current pen color.
        """
        return Color(self._state.color)

    def setcolor(self, value):
        """
        Modifies the pen color.

        Color can be specified as an (R, G, B) tuple or as a hex string or by
        name.
        """
        self._state.color = value

    def getfillcolor(self):
        """
        Return a tuple of (R, G, B) with the current fill color.
        """
        return Color(self._state.fillcolor)

    def setfillcolor(self, value):
        """
        Modifies the fill color.

        Color can be specified as an (R, G, B) tuple or as a hex string or by
        name.
        """
        self._state.fillcolor = value

    def getavatar(self):
        """
        Return a string with the name of the current avatar.
        """
        return self._state.avatar

    def setavatar(self, value):
        """
        Modifies the turtle avatar.
        """
        self._state.avatar = value

    def penup(self):
        """
        Raises the turtle pen so it stops drawing.

        Aliases: penup | pu
        """
        self._state.drawing = False

    def pendown(self):
        """
        Lower the turtle pen so it can draw in the screen.

        Aliases: pendown | pd
        """
        self._state.drawing = True

    def isdown(self):
        """
        Return True if the pen is down or False otherwise.
        """
        return self._state.drawing

    def isvisible(self):
        """
        Return True if the turtle is visible.
        """
        return not self._state.hidden

    def ishidden(self):
        """
        Return True if the turtle is not visible.
        """
        return self._state.hidden

    def hide(self):
        """
        Hide turtle.
        """
        self._state.hidden = True

    def show(self):
        """
        Shows a hidden turtle.
        """
        self._state.hidden = False

    def clean(self):
        """
        Clear all drawings made by turtle.
        """
        self._state.clean()

    def reset(self):
        """
        Clear all drawings and reset turtle to initial position.
        """
        self._state.reset()

    @vecargsmethod
    def goto(self, pos):
        """
        Goes to the given position.

        If the pen is down, it draws a line.
        """
        self._state.goto(pos)

    @vecargsmethod
    def jump(self, pos):
        """
        Relative movement by the desired position. It *never* draw as line even
        if the pen is down.
        """
        self._state.pos = pos + self._state.pos

    def forward(self, step):
        """
        Move the turtle forward by the given step size (in pixels).

        Aliases: forward | fd
        """
        self._state.step(step)

    def backward(self, step):
        """
        Move the turtle backward by the given step size (in pixels).

        Aliases: backward | back | bk
        """
        self.forward(-step)

    def left(self, angle):
        """
        Rotate the turtle counter-clockwise by the given angle.

        Aliases: left | lt

        Negative angles produces clockwise rotation.
        """
        self._state.rotate(angle)

    def right(self, angle):
        """
        Rotate the turtle clockwise by the given angle.

        Aliases: right | rt

        Negative angles produces counter-clockwise rotation. Return final
        heading.
        """
        self.left(-angle)

    # Aliases
    fd = forward
    bk = back = backward
    lt = left
    rt = right
    pu = penup
    pd = pendown

    # Attribute-based access to state
    pos = property(getpos, setpos)
    heading = property(getheading, setheading)
    width = property(getwidth, setwidth)
    color = property(getcolor, setcolor)
    fillcolor = property(getfillcolor, setfillcolor)
    drawing = property(isdown, lambda _, v: _.pendown() if v else _.penup())
    hidden = property(ishidden, lambda _, v: _.hide() if v else _.show())



import re

from pygments import unistring as uni
from pygments.lexer import bygroups, default, words
from pygments.lexers.python import Python3Lexer
from pygments.token import Text, Operator, Keyword, Name, String, Number
from pygments.util import shebang_matches


def transpyler_lexer_factory(transpyler):
    """
    Return a Pygments lexer class for the given transpyler.
    """

    def analyse_text(text):
        return shebang_matches(text, r'pythonw?3(\.\d)?')

    return type(
        transpyler.pygments_class_name,
        (Python3Lexer,),
        dict(
            analyse_text=analyse_text,
            name=transpyler.name,
            aliases=[transpyler.display_name],
            filenames=transpyler.file_extensions,
            mimetypes=transpyler.mimetypes,
            flags=re.MULTILINE | re.UNICODE,
            uni_name="[%s][%s]*" % (uni.xid_start, uni.xid_continue),
            tokens=make_transpyled_tokens(transpyler),
        )
    )


def make_transpyled_tokens(transpyler):
    """
    Return a list of pygments make_transpyled_tokens from a transpyler object.
    """

    keywords = transpyler.introspection.all_keywords
    constants = transpyler.introspection.all_constants
    exceptions = transpyler.introspection.all_exceptions
    builtins = transpyler.introspection.all_builtins

    uni_name = "[%s][%s]*" % (uni.xid_start, uni.xid_continue)

    tokens = Python3Lexer.tokens.copy()

    tokens['keywords'] = [
        (words(keywords, suffix=r'\b'), Keyword),
        (words(constants, suffix=r'\b'), Keyword.Constant),
    ]
    tokens['builtins'] = [
        (words(builtins, prefix=r'(?<!\.)', suffix=r'\b'), Name.Builtin),
        (r'(?<!\.)(self|Ellipsis|NotImplemented)\b', Name.Builtin.Pseudo),
        (words(exceptions, prefix=r'(?<!\.)', suffix=r'\b'),
         Name.Exception),
    ]
    tokens['numbers'] = [
        (r'(\d+\.\d*|\d*\.\d+)([eE][+-]?[0-9]+)?', Number.Float),
        (r'0[oO][0-7]+', Number.Oct),
        (r'0[bB][01]+', Number.Bin),
        (r'0[xX][a-fA-F0-9]+', Number.Hex),
        (r'\d+', Number.Integer)
    ]
    tokens['backtick'] = []
    tokens['name'] = [
        (r'@\w+', Name.Decorator),
        (uni_name, Name),
    ]
    tokens['funcname'] = [
        (uni_name, Name.Function, '#pop')
    ]
    tokens['classname'] = [
        (uni_name, Name.Class, '#pop')
    ]
    tokens['import'] = [
        (r'(\s+)(as)(\s+)', bygroups(Text, Keyword, Text)),
        (r'\.', Name.Namespace),
        (uni_name, Name.Namespace),
        (r'(\s*)(,)(\s*)', bygroups(Text, Operator, Text)),
        default('#pop')  # all else: go back
    ]
    tokens['fromimport'] = [
        (r'(\s+)(import)\b', bygroups(Text, Keyword), '#pop'),
        (r'\.', Name.Namespace),
        (uni_name, Name.Namespace),
        default('#pop'),
    ]
    tokens['strings'] = [
        # the old style '%s' % (...) string formatting (still valid in Py3)
        (r'%(\(\w+\))?[-#0 +]*([0-9]+|[*])?(\.([0-9]+|[*]))?'
         '[hlL]?[diouxXeEfFgGcrs%]', String.Interpol),
        # the new style '{}'.format(...) string formatting
        (r'\{'
         '((\w+)((\.\w+)|(\[[^\]]+\]))*)?'  # field name
         '(\![sra])?'  # conversion
         '(\:(.?[<>=\^])?[-+ ]?#?0?(\d+)?,?(\.\d+)?[bcdeEfFgGnosxX%]?)?'
         '\}', String.Interpol),
        # backslashes, quotes and formatting signs must be parsed one at a time
        (r'[^\\\'"%\{\n]+', String),
        (r'[\'"\\]', String),
        # unhandled string formatting sign
        (r'%|(\{{1,2})', String)
        # newlines are an error (use "nl" state)
    ]
    return tokens



import operator as op
from functools import wraps as _wraps

from transpyler.math import Vec as _Vec

COLORS = dict(
    preto='black', branco='white', vermelho='red', verde='green', azul='blue',
    amarelo='yellow',
)


def vecargsmethod(func):
    """
    Decorates a function of a vec object to accept the following signatures:

        func(vec, **kwds)
        func(x, y, **kwds)

    A Vec object is always passed to the given implementation.
    """

    @_wraps(func)
    def decorated(self, x, y=None, **kwds):
        if y is None:
            try:
                x, y = x
            except ValueError:
                raise ValueError('expected 2 elements, got %s' % len(x))

            return func(self, _Vec(x, y), **kwds)
        else:
            return func(self, _Vec(x, y), **kwds)

    return decorated


def alias(*args):
    """
    Set a list of function aliases for TurtleFunction methods.

    The aliases are automatically included in the resulting global_namespace.
    """

    def decorator(func):
        func.alias_list = args
        return func

    return decorator


def object_ctrl(getter, setter=None):
    def fget(self):
        return getattr(self.object, getter)()

    if setter is not None:
        def fset(self, value):
            getattr(self.object, setter)(value)

        return property(fget, fset)
    else:
        return property(fget)


def webcolor(color, source='pt_BR'):
    """
    Convert a color string back to a normalized english webcolor name.
    """

    color = color.lower().replace(' ', '')
    return COLORS.get(color, color)


def getsetter(name):
    """
    Return a property object that assigns that uses a (get*, set*) pair of
    methods.
    """

    getter = op.methodcaller('get' + name)
    setter = lambda self, v: setattr(self, 'set' + name, v)
    return property(getter, setter)


def turtle_property(getter, setter=None):
    """
    A property that uses getter/setters at .turtle.<getter> and .turtle.<setter>
    """

    fgetter = op.attrgetter('turtle.' + getter)
    fsetter = op.attrgetter('turtle.' + (setter or getter))

    def fget(x):
        return fgetter(x)()

    def fset(x, v):
        fsetter(x)(v)

    return property(fget, fset)


def ipc_property(name, unwrap=lambda x: x, wrap=lambda x: x):
    """
    A property that redirects requests to the getvalue/setvalue
    functions.
    """

    def fget(self):
        return unwrap(self.getvalue(name))

    def fset(self, value):
        self.setvalue(name, wrap(value))

    return property(fget, fset)



from collections.abc import Mapping

from .translate import translate_namespace, translator_factory
from .utils import extract_namespace, pretty_callable


class Namespace(Mapping):
    @property
    def namespace(self):
        if self._namespace is None:
            self._namespace = namespace(self.transpyler)
        return self._namespace

    def __init__(self, transpyler):
        self.transpyler = transpyler
        self._namespace = None

    def __len__(self):
        return len(self.namespace)

    def __iter__(self):
        return iter(self.namespace)

    def __getitem__(self, key):
        return self.namespace[key]


def namespace(transpyler):
    ns = global_functions(transpyler)

    # Load turtle functions
    if transpyler.has_turtle_functions and transpyler.turtle_backend:
        turtle_ns = turtle_functions(transpyler.turtle_backend)
        ns.update(turtle_ns)

    # Load special functions
    ns['exit'] = exit_function(transpyler.exit_callback, transpyler.translate)

    # Load default translations from using the lang option
    if transpyler.lang:
        translated = translate_namespace(ns, transpyler.lang)
        ns.update(translated)

    return ns


def global_functions(transpyler):
    """
    Return a dictionary with the default global namespace for the
    transpyler runtime.
    """
    from . import lib

    ns = extract_namespace(lib)
    return ns


def turtle_functions(backend):
    """
    Return a dictionary with all turtle-related functions.
    """

    if backend == 'tk':
        from .turtle.tk import make_turtle_namespace

        ns = make_turtle_namespace()
    elif backend == 'qt':
        from .turtle.qt import make_turtle_namespace

        ns = make_turtle_namespace()
    else:
        raise ValueError('invalid backend: %r' % backend)

    return ns


#
# Special functions
#
def exit_function(function, translate=translator_factory('en')):
    """
    Wraps the exit function into a callable object that prints a nice
    message for its repr.
    """

    @pretty_callable(translate('exit.doc'))
    def exit():
        return function()

    exit.__name__ = translate('exit.name')
    exit.__doc__ = translate('exit.doc')
    return exit



from collections.abc import Mapping
from functools import wraps

from .turtle import Turtle


class TurtleNamespace(Mapping):
    """
    Define all global functions that should be used in the turtle namespace.

    It behaves as an immutable mapping.
    """

    def __init__(self, turtle_cls=Turtle):
        self.turtle_cls = turtle_cls
        self._data = {}
        self._mainturtle = None
        self._maingroup = None
        self._data.update(self._namespace())

    def __getitem__(self, item):
        return self._data[item]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def mainturtle(self):
        """
        Returns the main turtle object.
        """

        if self._mainturtle is None:
            self._mainturtle = self.turtle_cls()
        return self._mainturtle

    def _from_turtle(self, method: callable):
        name = method.__name__

        @wraps(method)
        def func(*args, **kwargs):
            turtle = self.mainturtle()
            return getattr(turtle, name)(*args, **kwargs)

        return func

    def _namespace(self):
        wrap = self._from_turtle
        cls = self.turtle_cls
        return dict(
            mainturtle=self.mainturtle,
            Turtle=self.turtle_cls,

            # Getters and setters
            getpos=wrap(cls.getpos),
            setpos=wrap(cls.setpos),
            getheading=wrap(cls.getheading),
            setheading=wrap(cls.setheading),
            getwidth=wrap(cls.getwidth),
            setwidth=wrap(cls.setwidth),
            getcolor=wrap(cls.getcolor),
            setcolor=wrap(cls.setcolor),
            getfillcolor=wrap(cls.getfillcolor),
            setfillcolor=wrap(cls.setfillcolor),
            getavatar=wrap(cls.getavatar),
            setavatar=wrap(cls.setavatar),

            # Visibility
            penup=wrap(cls.penup),
            pendown=wrap(cls.pendown),
            isdown=wrap(cls.isdown),
            isvisible=wrap(cls.isvisible),
            ishidden=wrap(cls.ishidden),
            hide=wrap(cls.hide),
            show=wrap(cls.show),
            clean=wrap(cls.clean),
            reset=wrap(cls.reset),

            # Movement
            goto=wrap(cls.goto),
            jump=wrap(cls.jump),
            forward=wrap(cls.forward),
            backward=wrap(cls.backward),
            left=wrap(cls.left),
            right=wrap(cls.right),
            setspeed=self.setspeed,
            getspeed=self.getspeed,

            # Aliases
            fd=wrap(cls.forward),
            bk=wrap(cls.backward),
            back=wrap(cls.backward),
            lt=wrap(cls.left),
            rt=wrap(cls.right),
            pu=wrap(cls.penup),
            pd=wrap(cls.pendown),
        )

    def getspeed(self):
        return 5

    def setspeed(self, value):
        pass



import collections
from multiprocessing import Queue

from .state import TurtleState


class StateGroup(collections.MutableSequence):
    """
    Manage and coordinate a group of turtle states.

    It presents itself as a list of turtles.
    """

    state_class = TurtleState

    def __init__(self):
        self._index = 0
        self._turtles = []

    def __setitem__(self, idx, value):
        if not isinstance(value, self.state_class):
            raise TypeError(type(value))
        value.register(self)
        self._turtles.__setitem__(idx, value)

    def __len__(self):
        return len(self._turtles)

    def __iter__(self):
        return iter(self._turtles)

    def __getitem__(self, idx):
        return self._turtles[idx]

    def __delitem__(self, idx):
        return self._turtles.__delitem__[idx]

    def getturtle(self, id):
        """
        Gets turtle by id.
        """
        for turtle in self:
            if turtle.id == id:
                return turtle
        raise ValueError('invalid turtle id: %s' % id)

    def insert(self, idx, value):
        self._turtles.insert(idx, None)
        try:
            self[idx] = value
        except Exception:
            self.remove(None)

    def new_turtle(self, **kwargs):
        """
        Creates a new turtle state passing the given parameters.

        Return the turtle state object.
        """

        self._index += 1
        turtle = self.state_class(**kwargs)
        turtle.id = self._index
        self.append(turtle)
        return turtle

    def clean(self):
        """
        Clear all drawings, but maintains turtles in their respective states.
        """
        for turtle in self:
            turtle.clean()

    def reset(self):
        """
        Clear all drawings, and reset turtle state.
        """
        for turtle in self:
            turtle.reset()


class IpcStateGroup(StateGroup):
    """
    A state handle that receives messages from a client using a inbox/outbox
    communication model.
    """

    inbox_factory = outbox_factory = Queue
    timeout = 1.0
    inbox = outbox = None

    def __init__(self, inbox=None, outbox=None, **kwargs):
        self.inbox = inbox or self.inbox_factory()
        self.outbox = outbox or self.outbox_factory()
        super().__init__(**kwargs)

    def ping(self, receive=False):
        """
        Handles a ping message.
        """

        if receive:
            msg = self.inbox.get(timeout=self.timeout)
            if msg != ['ping']:
                raise ValueError('invalid ping message: %s' % msg)
            self.outbox.put(msg)
        else:
            self.outbox.put(['ping'])
            msg = self.inbox.get(timeout=self.timeout)
            if msg != ['ping']:
                raise ValueError('invalid ping message: %s' % msg)

    def recv(self):
        """
        Reads message from the inbox and sends response.
        """

        msg = self.inbox.get(timeout=self.timeout)
        reply = self.handle(msg)
        self.outbox.put(reply)
        return reply

    def handle(self, msg):
        """
        Receive message and handle message.

        Dispatch to the corresponding turtle, if required.
        """

        action, *args = msg

        # Global actions
        if action == 'newturtle':
            kwargs = args[0]
            new = self.new_turtle(**kwargs)
            return ['newturtle', new.id]
        elif action == 'clear':
            self.clean()
            return ['clear']
        elif action == 'reset':
            self.reset()
            return ['reset']

        # Turtle actions
        elif action in ('get', 'set', 'move', 'step', 'rotate'):
            id = args[0]
            turtle = self.getturtle(id)
            return turtle.handle([action] + args)
        else:
            raise ValueError('invalid action: %r' % action)



import sys
from multiprocessing import Process, Queue

from PyQt5 import QtWidgets

from .scene import TurtleScene
from .. import MailboxState


def start_qt_scene_app_subprocess():
    """
    Starts a remote sub-process that initializes a TurtleScene widget and Qt's
    mainloop.
    """

    inbox = MailboxState.inbox = Queue()
    outbox = MailboxState.outbox = Queue()
    process = Process(target=start_qt_scene_app,
                      kwargs=dict(outbox=outbox, inbox=inbox, ping=True),
                      name='turtle-server')
    process.daemon = True
    process.start()

    # Send a ping message to the out process
    outbox.put(['ping'])
    msg = inbox.get(timeout=2.0)
    if msg != ['ping']:
        raise RuntimeError('wrong response from server: %s' % (msg,))

    return process


def start_qt_scene_app(inbox=None, outbox=None, ping=False):
    """
    Starts a simple QtApp with a TurtleScene widget.

    Args:
        inbox/outbox:
            Inbox/Outbox queues used for IPC.
    """
    from .view import TurtleView

    app = QtWidgets.QApplication(sys.argv)
    scene = TurtleScene(inbox=inbox, outbox=outbox)
    window = TurtleView(scene)
    window.setWindowTitle('Turtle')
    window.setMinimumWidth(800)
    window.setMinimumHeight(600)
    if ping:
        scene.ping(receive=True)
    window.show()
    sys.exit(app.exec_())


def start_console():
    """
    Starts a console with the Qt interface.
    """

    raise NotImplementedError('Qt interface does not support console app yet!')



"""
Useful mathematical functions and objects.

These functions are used mostly by the turtle interface.
"""

import math as _math
import random as _random
from collections import namedtuple as _namedtuple
from numbers import Number, Real

_dg = _math.pi / 180
pi = _math.pi
e = _math.e


# Trigonometric functions
def cos(angle):
    """Cosine of an angle (in degrees)"""

    angle = angle % 360
    if angle in (90, 270):
        return 0.0
    return _math.cos(angle * _dg)


def sin(angle):
    """Sine of an angle (in degrees)"""

    angle = angle % 360
    if angle in (0, 180):
        return 0.0
    return _math.sin(angle * _dg)


def tan(angle):
    """Tangent of an angle (in degrees)"""

    angle = angle % 180
    if angle == 45:
        return 1.0
    return _math.tan(angle * _dg)


# Other functions
def sqrt(x: Real) -> Real:
    """
    Return the square root of a positive number.
    """
    return _math.sqrt(x)


def exp(x: Real) -> Real:
    """
    Return the exponential of a number.
    """
    return _math.exp(x)


def log(x: Real) -> Real:
    """
    Return the natural logarithm of x.

    Aliases: log | ln
    """
    return _math.log(x)


ln = log


def log10(x: Real) -> Real:
    """
    Return the logarithm of x in base 10.
    """
    return _math.log10(x)


def log2(x: Real) -> Real:
    """
    Return the logarithm of x in base 2.
    """
    return _math.log2(x)


def sign(x):
    """
    Return 1 if x is positive, -1 if it is negative and 0 if it is null.

    Examples:
        >>> sign(-32.0)
        -1
    """
    try:
        if x == 0:
            return 0
        elif x > 0:
            return 1
        elif x < 0:
            return -1
    except TypeError:
        pass
    raise ValueError(sign.ERROR)


sign.ERROR = 'argument does not have a well defined sign'

# Rounding
_abs, _round = abs, round


def abs(x) -> Number:
    """
    Return the absolute value of the argument.
    """
    return _abs(x)


def round(number: Number, ndigits=None) -> Number:
    """
    Round a number to a given precision in decimal digits (default 0 digits).
    ndigits may be negative.

    Examples:
        >>> round(1.9)
        2
        >>> round(3.141516, 2)
        3.14
    """
    if ndigits is None:
        return _round(number)
    return _round(number, ndigits)


def ceil(number: Real) -> int:
    """
    Return the ceiling of x as an Integral.
    This is the smallest integer >= x.
    """
    return _math.ceil(number)


def trunc(number: Real) -> int:
    """
    Truncates x to the nearest Integral toward 0.
    """
    return _math.trunc(number)


# Operations on collections of numbers
_min, _max, _sum = min, max, sum


def sum(seq, start=0.0):
    """
    Return the sum of all numbers in the sequence.
    """
    return _sum(seq, start)


def product(seq, start=1.0):
    """
    Return the product of all numbers in the sequence.
    """

    result = start
    for x in seq:
        result *= x
    return result


def min(*args, **kwargs):
    """
    Return the minimum value in the arguments.

    It called with a single sequence, return the minimum value on that sequence.
    """
    return _min(*args)


def max(*args, **kwargs):
    """
    Return the maximum value in the arguments.

    It called with a single sequence, return the maximum value on that sequence.
    """
    return _max(*args)


# Random numbers
def dice(n=6):
    """
    Simulate a dice trow (return a random number between 1 and 6.

    Call with an explicit argument to throw a dice with a different number of
    faces

    Examples:
        >>> dice(20)                                            # doctest: +SKIP
        13
    """
    return _random.randint(1, 6)


def random():
    """
    Return a random number between 0 and 1.
    """
    return _random.random()


def randint(a, b):
    """
    Return a random integer between a and b (inclusive).
    """
    return _random.randint(a, b)


# Vectors and linear algebra
class Vec(_namedtuple('Vec', ['x', 'y'])):
    """
    A tuple-based 2D vector.

    Supports all basic arithmetic operations.
    """

    @classmethod
    def from_angle(cls, angle, length=1):
        """
        Creates vector from angle and length.
        """
        return cls(cos(angle), sin(angle))

    def __new__(cls, x, y):
        return super(Vec, cls).__new__(cls, x + 0., y + 0.)

    def __repr__(self):
        return '(%g, %g)' % (self.x, self.y)

    def __add__(self, other):
        x, y = other
        return Vec(x + self.x, y + self.y)

    def __sub__(self, other):
        x, y = other
        return Vec(self.x - x, self.y - y)

    def __mul__(self, other):
        return Vec(other * self.x, other * self.y)

    def __truediv__(self, other):
        return Vec(self.x / other, self.y / other)

    def __radd__(self, other):
        return self + other

    def __rsub__(self, other):
        return self * (-1) + other

    def __rmul__(self, other):
        return self * other

    def __neg__(self):
        return Vec(-self.x, -self.y)

    def __abs__(self):
        return _math.sqrt(self.x ** 2 + self.y ** 2)

    def norm(self):
        """
        Vector norm.
        """
        return self.__abs__()

    def normalized(self):
        """
        Return unity vector.
        """
        return self / abs(self)

    def perp(self, invert=False):
        """
        Returns a perpendicular vector rotated 90 degrees counter-clockwise.
        """

        if invert:
            return Vec(self.y, -self.x)
        return Vec(-self.y, self.x)

    def rotate(self, angle):
        """
        Return rotated vector by the given angle.
        """

        x, y = self
        c, s = cos(angle), sin(angle)
        return Vec(x * c - y * s, x * s + y * c)


def vec(x, y=None):
    """
    Return a vector with the given (x, y) components.
    """

    if y is None:
        x, y = x
    return Vec(x, y)


def dot(u, v):
    """
    The dot product (scalar product) of two vectors.
    """
    return _sum(x * y for (x, y) in zip(u, v))



import builtins as _builtins
import codeop

from lazyutils import lazy

from .info import Info
from .introspection import Introspection
from .lexer import Lexer
from .namespace import Namespace
from .translate import translator_factory
from .utils.utils import has_qt

# Save useful builtin functions
_compile = _builtins.compile
_exec = _builtins.exec
_eval = _builtins.eval
_input = _builtins.input
_print = _builtins.print


class SingletonMeta(type):
    """
    Base metaclass for classes that have a single instance.
    """

    _subclasses = []

    def __init__(cls, *args, **kwargs):  # noqa: N805
        super().__init__(*args, **kwargs)
        type(cls)._subclasses.append(cls)

    def __call__(cls, *args, **kwargs):  # noqa: N805
        try:
            return cls._instance
        except AttributeError:
            Transpyler._instance = super().__call__(*args, **kwargs)
            return cls._instance


class Transpyler(metaclass=SingletonMeta):
    """
    Base class for all new Transpylers.

    A transpyler is a singleton object.

    Very simple Python variations can be created by subclassing Transpyler::

        class PyBr(Transpyler):
            translations = {
                'para': 'for',            # single token translations
                'em': 'in',
                ('para', 'cada'): 'for',  # token sequence translations
                ('faça', ':'): ':',
            }

    Now we can create an object with exec(), eval() and compile() functions that
    handle the newly defined transpyler:

        pybr = PyBr()
        global_ns = {}

        pybr.exec('''
        x, y = 1, 1
        para cada i em [1, 2, 3, 4, 5] faça:
            x, y = y, x + y
        ''', global_ns)

        assert globals_ns['x'] == 8
        assert globals_ns['y'] == 13
    """

    # Cache builtins
    _compile = _compile
    _exec = _exec
    _eval = _eval
    _input = _input
    _print = _print

    # Factories and subclasses
    lexer_factory = Lexer
    info_factory = Info
    introspection_factory = Introspection
    namespace_factory = Namespace

    # Constants
    lang = 'en'
    has_turtle_functions = False
    turtle_backend = None
    standard_lib = None
    translations = None
    invalid_tokens = None
    language_version = '0.1.0'
    version = '0.1.0'
    codemirror_mode = 'python'
    file_extension = 'py'

    # Language info and introspection
    introspection = lazy(lambda self: self.introspection_factory(self))
    info = lazy(lambda self: self.info_factory(self))
    mimetypes = lazy(lambda self: [self.mimetype])
    mimetype = lazy(lambda self: 'text/x-%s' % self.name)
    link_docs = lazy(
        lambda self:
        "http://%s.readthedocs.io/%s/latest/" % (self.name, self.lang)
    )
    link_github = lazy(
        lambda self: "http://github.com/transpyler/%s/" % self.name
    )
    translate = lazy(lambda self: translator_factory(self.lang))

    # Display messages
    display_name = lazy(lambda self: self.name.title().replace('_', ' '))
    short_banner = lazy(
        lambda self: self.translate(
            '%s %s\n'
            'Type "help", "copyright" or "license" for more information.' %
            (self.display_name, self.version))
    )
    long_banner = lazy(lambda self: self.short_banner)
    use_short_banner = True

    # Lexer
    lexer = lazy(lambda self: self.lexer_factory(self))

    @lazy
    def name(self):
        cls_name = self.__class__.__name__.lower()
        if cls_name == 'transpyler':
            return 'transpyler'
        elif cls_name.endswith('transpyler'):
            return cls_name[:-10]
        else:
            return cls_name

    @lazy
    def namespace(self):
        return self.recreate_namespace()

    def __init__(self, **kwargs):
        self._forbidden = False
        for k, v in kwargs.items():
            setattr(self, k, v)
        self._has_init = False

        assert self.name, 'Name cannot be empty'

    def __repr__(self):
        return '<%s: %r>' % (self.__class__.__name__, self.name)

    #
    #  System functions
    #
    def init(self, ns=None):
        """
        Initializes transpyler runtime.

        Args:
            ns (dict):
                A dictionary with extra functions to be added to the
                globals namespace at runtime.
        """

        self.apply_curses()
        self.namespace.update(ns or {})

    def apply_curses(self):
        """
        Apply any required curses.

        Default implementation does nothing.
        """

    def compile(self, source, filename, mode, flags=0, dont_inherit=False,
                compile_function=None):
        """
        Similar to the built-in function compile() for Transpyled code.

        The additional compile_function() argument allows to define a function
        to replace Python's builtin compile().

        Args:
            source (str or code):
                Code to be executed.
            filename:
                File name associated with code. Use '<input>' for strings.
            mode:
                One of 'exec' or 'eval'. The second allows only simple
                statements that generate a value and is used by the eval()
                function.
            forbidden (bool):
                If true, initialize the forbidden lib functionality to enable
                i18n for Python builtins in C-level.
            compile_function (callable):
                A possible replacement for Python's built-in compile().
        """

        compile_function = compile_function or _compile
        source = self.transpile(source)
        return compile_function(source, filename, mode, flags, dont_inherit)

    def exec(self, source, globals=None, locals=None, exec_function=None):
        """
        Similar to the built-in function exec() for transpyled code.

        The additional exec_function() argument allows to define a function to
        replace Python's builtin compile().

        Args:
            source (str or code):
                Code to be executed.
            globals, locals:
                A globals/locals dictionary
            exec_function (callable):
                A possible replacement for Python's built-in exec().
        """

        exec_function = exec_function or _exec
        code = self.transpile(source) if isinstance(source, str) else source
        globals = {} if globals is None else globals
        globals.update(self.namespace)

        args = (globals,) if locals is None else (globals, locals)
        return exec_function(code, *args)

    def eval(self, source, globals=None, locals=None, eval_function=None):
        """
        Similar to the built-in function eval() for transpyled code.

        The additional eval_function() argument allows to define a function to
        replace Python's builtin compile().

        Args:
            source (str or code):
                Code to be executed.
            globals, locals:
                A globals/locals dictionary
            eval_function (callable):
                A possible replacement for Python's built-in eval().
        """
        eval_function = eval_function or _eval
        code = self.transpile(source) if isinstance(source, str) else source
        globals = {} if globals is None else globals
        globals.update(self.namespace)

        args = (globals,) if locals is None else (globals, locals)
        return eval_function(code, *args)

    def transpile(self, src):
        """
        Convert source to Python.
        """

        return self.lexer.transpile(src)

    def is_incomplete_source(self, src, filename="<input>", symbol="single"):
        """
        Test if a given source code is incomplete.

        Incomplete code may appear in users interactions when user is typing a
        multi line command:

        for x in range(10):
            ... should continue here, but user already pressed enter!
        """

        try:
            pytuga_src = self.transpile(src)
        except SyntaxError:
            return True
        return codeop.compile_command(pytuga_src, filename, symbol) is None

    @classmethod  # noqa: C901 (it only creates functions on a closure)
    def core_functions(cls):
        """
        Return an dictionary with a small namespace for the core functions in
        the transpyler API:

        * init: init runtime
        * compile: compile a string of source code
        * exec: execute a string of source code
        * eval: evaluate a string of source code and return the resulting object
        * transpile: transpile source code to Python
        * namespace: return a dictionary with builtin functions
        * is_complete_source: check if string can be executed as-is or if it
        requires additional lines of code in order to execute.
        """

        def init(ns=None):
            return cls().init(ns)

        def compile(source, filename, mode, flags=0, dont_inherit=False,
                    compile_function=None):
            return cls().compile(
                source, filename, mode, flags=flags,
                dont_inherit=dont_inherit, compile_function=compile_function
            )

        def exec(source, globals=None, locals=None, exec_function=None):
            return cls().exec(
                source, globals=globals, locals=locals,
                exec_function=exec_function,
            )

        def eval(source, globals=None, locals=None, eval_function=None):
            return cls().eval(
                source, globals=globals, locals=locals,
                eval_function=eval_function,
            )

        def transpile(src):
            return cls().transpile(src)

        def is_incomplete_source(src, filename="<input>", symbol="single"):
            return cls().is_incomplete_source(src, filename, symbol)

        def namespace(turtle=None):
            """
            Return a dictionary with all public functions.

            If turtle is given and it is either 'qt' or 'tk', it includes the
            corresponding turtle functions into the namespace.
            """
            transpyler = cls()
            transpyler.has_turtle_functions = turtle is not None
            transpyler.turtle_backend = turtle
            transpyler.init()
            return transpyler.namespace

        # Update docstrings
        init.__doc__ = cls.init.__doc__
        compile.__doc__ = cls.compile.__doc__
        exec.__doc__ = cls.exec.__doc__
        eval.__doc__ = cls.eval.__doc__
        transpile.__doc__ = cls.transpile.__doc__
        is_incomplete_source.__doc__ = cls.is_incomplete_source.__doc__

        return dict(
            init=init, compile=compile, exec=exec, eval=eval,
            transpile=transpile, is_incomplete_source=is_incomplete_source,
            namespace=namespace,
        )

    #
    # Console helpers
    #
    def console_banner(self, short=None):
        """
        Return a string with the console banner.
        """
        if short is None:
            short = self.use_short_banner

        if short:
            return self.short_banner
        return getattr(self, 'banner', self.short_banner)

    def recreate_namespace(self):
        """
        Recompute the default namespace for the transpyler object.
        """
        ns = self.namespace_factory(self)
        self.namespace = dict(ns)
        return self.namespace

    #
    # External execution
    #
    def start_console(self, console='auto'):
        """
        Starts a regular python console with the current transpyler.

        Args:
            console:
                Can be one of 'jupyter', 'console', 'qtconsole', 'auto'. This
                chooses the default console application. The default behavior
                (auto) is to try jupyter and fallback to console if it is
                not available.
        """

        # Select the console application
        if console == 'auto':
            try:
                import IPython  # noqa: F401
            except ImportError:
                console = 'console'
            else:
                console = 'jupyter'

        if console == 'qtconsole':
            from .jupyter import start_jupyter
            start_jupyter(transpyler=self, gui=True)

        elif console == 'jupyter':
            from .jupyter import start_jupyter
            start_jupyter(transpyler=self, gui=False)

        elif console == 'console':
            from .console import start_console
            start_console(transpyler=self)

        else:
            raise ValueError('invalid console: %r' % console)

    def start_notebook(self):
        """
        Starts a jupyter notebook with the current transpyler.
        """

        from .jupyter import start_notebook
        start_notebook(self)

    def start_qturtle(self):
        """
        Starts a QTurtle application with the current transpyler.
        """

        if not has_qt():
            raise SystemExit('PyQt5 is necessary to run the turtle '
                             'application.')

        from qturtle.mainwindow import start_application
        start_application(self)

    def start_main(self):
        """
        Starts the default main application.
        """

        import click

        @click.command()
        @click.option('--cli', '-c', is_flag=True, default=False,
                      help='start gui-less console.')
        @click.option('--console', is_flag=True, default=False,
                      help='start a simple gui-less console.')
        @click.option('--notebook/--no-notebook', '-n', default=False,
                      help='starts notebook server.')
        def main(cli, notebook, console):
            if cli:
                return self.start_console('auto')
            if console:
                return self.start_console('console')
            if notebook:
                return self.start_notebook()

            if has_qt():
                return self.start_qturtle()
            else:
                msg = 'Could not start GUI. Do you have Qt installed?'
                click.echo(msg, err=True)
                return self.start_console('jupyter')

        return main()

    #
    # Callbacks: those methods are designed to be overridden by instances
    #
    def exit_callback(self):
        print('bye!')
        raise SystemExit(0)


#
# Utility functions
#
def get_transpyler() -> Transpyler:
    """
    Return an instance of the last defined transpyler class.
    """

    return SingletonMeta._subclasses[-1]()


def extract_namespace(mod):
    """
    Return a dictionary with module public variables.
    """

    return {
        name: getattr(mod, name)
        for name in dir(mod) if not name.startswith('_')
    }



import codecs
import os
import re

from setuptools import setup, find_packages

# Extract version
init = open(os.path.join('src', 'transpyler', '__init__.py')).read()
m = re.search(r"__version__ ?= ?'([0-9a-z.]+)'", init)
version = m.group(1)

# Reuse dependencies list
dependencies = {
    'google_translate': ['textblob'],
    'jupyter': ['jupyter', 'jupyter-console', 'ipython'],
    'pygments': ['pygments'],
    'dev': ['manuel', 'pytest', 'pytest-cov']
}
dependencies['dev'] += sum(
    [v for k, v in dependencies.items() if k != 'dev'], []
)

setup(
    # Basic info
    name='transpyler',
    version=version,
    author='Fábio Macêdo Mendes',
    author_email='fabiomacedomendes@gmail.com',
    url='https://github.com/transpyler/transpyler/',
    description='A framework for building localized Python-like languages.',
    long_description=codecs.open('README.rst', 'rb', 'utf8').read(),

    # Classifiers (see https://pypi.python.org/pypi?%3Aaction=list_classifiers)
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
    ],

    # Package data
    package_data={
        'transpyler.jupyter': [
            'assets/*.*',
        ],
        'transpyler.turtle': [
            'data/*.*',
        ],
        'transpyler.l10n': [
            '*.po',
            '*.mo',
            '*.pot',
        ],
    },

    # Packages and dependencies
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=[
        'lazyutils',
        'unidecode',
        'polib',
        'colortools',
        'click',
    ],
    extras_require=dependencies,

    # Other configurations
    zip_safe=False,
    platforms='any',
)



"""
A turtle library with a client/server architecture.

The turtle client is responsible for issuing commands that are serialized and
sent to a turtle server. The server stores the state of the turtles and
can send messages the tell the client how to update its display.

In the simple case, both the server and the client run on the same machine and
"message passing" is a simple method call. The client and server can live on
different processes, different machines and can even be implemented in different
languages.

We have a few use cases:

* Implement turtle commands in a Jupyter architecture: the server runs in the
  kernel process and the client runs in the shell process.
* Jupyter notebooks: similar as above, but messages can be redirect to
  javascript to update a canvas object that display the different turtles.
* It should be possible to choose different frontend toolkits. Tk does not
  capture the main thread, but Qt does. We would like to support both
  archtectures.
* The "draw" messages can be stored and saved in a persistent structure. This
  can be used to generate output in static file formats or even to implement an
  online judge for turtle-based programs.


.. autoclass:: transpyler.turtle.Turtle
    :members:

.. autoclass:: transpyler.turtle.State
    :members:
"""

from .turtle import Turtle
from .turtlegroup import TurtleGroup
from .namespace import TurtleNamespace
from .state import TurtleState, MailboxState, MirrorState, RemoteState, PropertyState
from .stategroup import StateGroup, IpcStateGroup



from ipykernel.zmqshell import ZMQInteractiveShell
from lazyutils import lazy


class TranspylerShell(ZMQInteractiveShell):
    """
    A IPython based shell for transpyled languages.

    A shell is used by the kernel to interact with code.
    """

    @lazy
    def transpyler(self):
        return self.parent.transpyler

    def init_user_ns(self):
        super().init_user_ns()
        self.transpyler.exit_callback = self.exiter

    def init_create_namespaces(self, user_module=None, user_ns=None):
        super().init_create_namespaces(user_module, user_ns)

        ns = self.transpyler.namespace
        self.user_ns.update(ns)
        self.user_global_ns.update(ns)
        self.user_module.__dict__.update(ns)

    def ex(self, cmd):
        return super().ex(self.transpyler.transpile(cmd))

    def ev(self, cmd):
        return super().ev(self.transpyler.transpile(cmd))



import os

from PyQt5 import QtGui
from PyQt5 import QtSvg

from .utils import qtproperty, from_qvector, to_qvector, from_qcolor, to_qcolor
from .. import TurtleState, MailboxState, IpcStateGroup, Turtle as BaseTurtle, \
    TurtleNamespace

dir_path = os.path.dirname(os.path.dirname(__file__))
svg_path = os.path.join(dir_path, 'data', 'turtleart.svg')
svg_renderer = QtSvg.QSvgRenderer(svg_path)


class QGraphicsItemState(TurtleState):
    """
    A server turtle that redirects operations to a QGraphicsSvgItem.

    This turtle lives on the server part of the application.
    """

    valid_avatars = ['tuga']
    pos = qtproperty('graphics_item.pos', from_qvector, to_qvector)
    heading = qtproperty('graphics_item.rotation')
    width = qtproperty('pen.width')
    color = qtproperty('pen.color', from_qcolor, to_qcolor)
    fillcolor = qtproperty('brush.color', from_qcolor, to_qcolor)

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('width', 2)
        avatar = kwargs.setdefault('avatar', 'tuga')
        self.size = kwargs.pop('size', 45)
        self.validate_avatar(avatar)
        self.graphics_item = cursor = QtSvg.QGraphicsSvgItem()

        # Loads from turtleart.svg
        cursor.setSharedRenderer(svg_renderer)
        cursor.setElementId(avatar)

        # Define local transforms
        rect = cursor.sceneBoundingRect()
        curr_width, curr_height = rect.width(), rect.height()
        cursor.setTransform(QtGui.QTransform(
            1.00, 0.00,
            0.00, 1.00,
            -curr_width / 2, -curr_height / 2)
        )
        cursor.setTransformOriginPoint(0.5 * curr_width, 0.5 * curr_height)
        cursor.setScale(self.size / curr_width)
        cursor.setZValue(1.0)
        self.pen = QtGui.QPen(QtGui.QColor(0, 0, 0))
        self.brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        super().__init__(*args, **kwargs)

    def draw_line(self, v1, v2):
        a, b = v1
        c, d = v2
        line = self.group.scene.addLine(a, b, c, d, self.pen)
        self.lines.append(line)

    def clean(self):
        scene = self.group.scene
        lines = self.lines
        while lines:
            line = lines.pop()
            scene.removeItem(line)

    def register(self, group):
        super().register(group)
        group.scene.addItem(self.graphics_item)


class QGraphicsSceneGroup(IpcStateGroup):
    """
    A manager of turtle states.
    """

    state_class = QGraphicsItemState

    def __init__(self, scene, **kwargs):
        self.scene = scene
        super().__init__(**kwargs)


#
# This is the user-visible turtle class for the QtSceneGraph. IPC is done using
# two message queues. The client publishes messages in a queue that is consumed
# by the server and vice-versa.
#
class Turtle(BaseTurtle):
    """
    Creates a new Turtle.
    """
    _state_factory = MailboxState


def make_turtle_namespace():
    """
    Returns a dictionary with the namespace of turtle functions.
    """

    return dict(TurtleNamespace(Turtle))



import gettext as _gettext
import os

import polib

L10N_PATH = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'l10n')
LANGUAGE = None
GETTEXT_INSTANCE = None
PO_HEADER = """#
msgid ""
msgstr ""
"Project-Id-Version: \\n"
"Language: %s\\n"
"MIME-Version: 1.0\\n"
"Content-Type: text/plain; charset=UTF-8\\n"
"Content-Transfer-Encoding: 8bit\\n"

"""


def gettext(x):
    """
    Translates string using the default language.
    """
    if LANGUAGE is None:
        return x
    return GETTEXT_INSTANCE.gettext(x)


def set_language(lang):
    """
    Sets the global output language for the transpyler environment.
    """
    global LANGUAGE, GETTEXT_INSTANCE

    LANGUAGE = lang.replace('-', '_')
    GETTEXT_INSTANCE = gettext_for(LANGUAGE)


def gettext_for(lang):
    """
    Return a GNUTranslation class for the given language.

    Example:
        >>> trans = gettext_for('pt_BR')
        >>> _ = trans.gettext
        >>> _('hello world!')                                   # doctest: +SKIP
        'olá mundo!'
    """
    lang = lang.replace('-', '_')

    try:
        with open(os.path.join(L10N_PATH, lang + '.mo'), 'rb') as F:
            result = _gettext.GNUTranslations(F)
    except FileNotFoundError:
        result = _gettext.NullTranslations()
    return result


def create_pot_file(translations, path=None):
    """
    Save contents on a POT file from a dictionary of (id, string) translation
    pairs.

    It interprets the id as a comment in the POT file and the string is the
    msgid field. The msgstr field is always empty.
    """

    pot = polib.POFile()
    pot.metadata = {
        'Project-Id-Version': '1.0',
        'Report-Msgid-Bugs-To': 'fabiomacedomendes@gmail.com',
        'MIME-Version': '1.0',
        'Content-Type': 'text/plain; charset=utf-8',
        'Content-Transfer-Encoding': '8bit',
    }

    # We keep a dictionary from msgids's to comments in order to keep all
    # entries unique. Duplicate entries simply duplicate comments
    entries = {}
    for key, data in translations.items():
        entry = polib.POEntry(msgid=data, comment=key)
        if data in entries:
            entries[data].comment += '\n' + key
        else:
            entries[data] = entry
            pot.append(entry)
    if path:
        pot.save(path)
    return pot



from unidecode import unidecode


SYNONYM_ERROR_MSG = '%s is present in global_namespace, but is also a synonym of %s'


def collect_synonyms(namespace, add_unaccented=True):
    """
    Return a dictionary with all synonyms found in the given global_namespace.

    Args:
        namespace:
             A mapping from names to values.
        add_unaccented:
            If True (default) extends with the unaccented versions of the names
            collected in the global_namespace.

    Raise a ValueError for name conflicts.
    """

    result = {}

    # Includes the aliased versions of the names defined in the input
    # global_namespace
    for attr, func in namespace.items():
        if hasattr(func, '__synonyms__'):
            for alias in func.__synonyms__:
                result.setdefault(alias, func)

    # Collect unaccented names and maps them to the corresponding
    # functions/values
    if add_unaccented:
        for name, func in list(result.items()):
            no_accent = unidecode(name)
            if no_accent != name:
                result.setdefault(no_accent, func)

    return result


def extract_namespace(mod):
    """
    Return a namespace dict with all public names for the given module.

    If no module is given, uses transpyler's standard lib.
    """
    namespace = vars(mod)
    return {k: v for (k, v) in namespace.items() if not k.startswith('_')}


def full_class_name(cls):
    """
    Return the full class name prepending module paths.
    """

    return '%s.%s' % (cls.__module__, cls.__name__)
